#!/usr/bin/env python

# This script expects one line for each new revision on the form
#   <oldrev> <newrev> <refname>
#
# For example:
#   aa453216d1b3e49e7f6f98441fa56946ddcd6a20
#   68f7abf4e6f922807889f52bc043ecd31b79f814 refs/heads/master
#
# Each of these changes will be passed to the buildbot server along
# with any other change information we manage to extract from the
# repository.
#
# This script is meant to be run from hooks/post-receive in the git
# repository. It can also be run at client side with hooks/post-merge
# after using this wrapper:

# !/bin/sh
# PRE=$(git rev-parse 'HEAD@{1}')
# POST=$(git rev-parse HEAD)
# SYMNAME=$(git rev-parse --symbolic-full-name HEAD)
# echo "$PRE $POST $SYMNAME" | git_buildbot.py
#
# Largely based on contrib/hooks/post-receive-email from git.

from future.utils import iteritems

try:
    from future.utils import text_type
except ImportError:
    from six import text_type

import logging
import re
import shlex
import subprocess
import sys
from optparse import OptionParser

from twisted.cred import credentials
from twisted.internet import defer
from twisted.internet import reactor

try:
    from twisted.spread import pb
except ImportError as e:
    raise ImportError(
        'Twisted version conflicts.\
                      Upgrade to latest version may solve this problem.\
                      try:  pip install --upgrade twisted'
    ) from e

# Modify this to fit your setup, or pass in --master server:port on the
# command line

master = "localhost:9989"

# When sending the notification, send this category if (and only if)
# it's set (via --category)

category = None

# When sending the notification, send this repository if (and only if)
# it's set (via --repository)

repository = None

# When sending the notification, send this project if (and only if)
# it's set (via --project)

project = None

# When sending the notification, send this codebase.  If this is None, no
# codebase will be sent.  This can also be set via --codebase

codebase = None

# Username portion of PB login credentials to send the changes to the master
username = "change"

# Password portion of PB login credentials to send the changes to the master
auth = "changepw"

# When converting strings to unicode, assume this encoding.
# (set with --encoding)

encoding = sys.stdout.encoding or 'utf-8'

# If true, takes only the first parent commits. This controls if we want to
# trigger builds for merged in commits (when False).

first_parent = False

# The GIT_DIR environment variable must have been set up so that any
# git commands that are executed will operate on the repository we're
# installed in.

changes = []


def connectFailed(error):
    logging.error("Could not connect to %s: %s", master, error.getErrorMessage())
    return error


def addChanges(remote, changei, src='git'):
    logging.debug("addChanges %s, %s", repr(remote), repr(changei))

    def addChange(c):
        logging.info("New revision: %s", c['revision'][:8])
        for key, value in iteritems(c):
            logging.debug("  %s: %s", key, value)

        c['src'] = src
        d = remote.callRemote('addChange', c)
        return d

    finished_d = defer.Deferred()

    def iter():
        try:
            c = next(changei)
            logging.info("CHANGE: %s", c)
            d = addChange(c)
            # handle successful completion by re-iterating, but not immediately
            # as that will blow out the Python stack

            def cb(_):
                reactor.callLater(0, iter)

            d.addCallback(cb)
            # and pass errors along to the outer deferred
            d.addErrback(finished_d.errback)
        except StopIteration:
            remote.broker.transport.loseConnection()
            finished_d.callback(None)
        except Exception as e:
            logging.error(e)

    iter()
    return finished_d


def connected(remote):
    return addChanges(remote, changes.__iter__())


def grab_commit_info(c, rev):
    # Extract information about committer and files using git show
    options = "--raw --pretty=full"
    if first_parent:
        # Show the full diff for merges to avoid losing changes
        # when builds are not triggered for merged in commits
        options += " --diff-merges=first-parent"
    f = subprocess.Popen(shlex.split(f"git show {options} {rev}"), stdout=subprocess.PIPE)

    files = []
    comments = []

    while True:
        line = f.stdout.readline().decode(encoding)
        if not line:
            break

        if line.startswith(4 * ' '):
            comments.append(line[4:])

        m = re.match(r"^:.*[MAD]\s+(.+)$", line)
        if m:
            logging.debug("Got file: %s", m.group(1))
            files.append(text_type(m.group(1)))
            continue

        m = re.match(r"^Author:\s+(.+)$", line)
        if m:
            logging.debug("Got author: %s", m.group(1))
            c['who'] = text_type(m.group(1))

        # Retain default behavior if all commits trigger builds
        if not first_parent and re.match(r"^Merge: .*$", line):
            files.append('merge')

    c['comments'] = ''.join(comments)
    c['files'] = files
    status = f.wait()
    if status:
        logging.warning("git show exited with status %d", status)


def gen_changes(input, branch):
    while True:
        line = input.stdout.readline().decode(encoding)
        if not line:
            break

        logging.debug("Change: %s", line)

        m = re.match(r"^([0-9a-f]+) (.*)$", line.strip())
        c = {
            'revision': m.group(1),
            'branch': text_type(branch),
        }

        if category:
            c['category'] = text_type(category)

        if repository:
            c['repository'] = text_type(repository)

        if project:
            c['project'] = text_type(project)

        if codebase:
            c['codebase'] = text_type(codebase)

        grab_commit_info(c, m.group(1))
        changes.append(c)


def gen_create_branch_changes(newrev, refname, branch):
    # A new branch has been created. Generate changes for everything
    # up to `newrev' which does not exist in any branch but `refname'.
    #
    # Note that this may be inaccurate if two new branches are created
    # at the same time, pointing to the same commit, or if there are
    # commits that only exists in a common subset of the new branches.

    logging.info("Branch `%s' created", branch)

    p = subprocess.Popen(shlex.split(f"git rev-parse {refname}"), stdout=subprocess.PIPE)
    branchref = p.communicate()[0].strip().decode(encoding)
    f = subprocess.Popen(shlex.split("git rev-parse --not --branches"), stdout=subprocess.PIPE)
    f2 = subprocess.Popen(
        shlex.split(f"grep -v {branchref}"), stdin=f.stdout, stdout=subprocess.PIPE
    )
    options = "--reverse --pretty=oneline --stdin"
    if first_parent:
        # Don't add merged commits to avoid running builds twice for the same
        # changes, as they should only be done for first parent commits
        options += " --first-parent"
    f3 = subprocess.Popen(
        shlex.split(f"git rev-list {options} {newrev}"),
        stdin=f2.stdout,
        stdout=subprocess.PIPE,
    )

    gen_changes(f3, branch)

    status = f3.wait()
    if status:
        logging.warning("git rev-list exited with status %d", status)


def gen_create_tag_changes(newrev, refname, tag):
    # A new tag has been created. Generate one change for the commit
    # a tag may or may not coincide with the head of a branch, so
    # the "branch" attribute will hold the tag name.

    logging.info("Tag `%s' created", tag)
    f = subprocess.Popen(
        shlex.split(f"git log -n 1 --pretty=oneline {newrev}"), stdout=subprocess.PIPE
    )
    gen_changes(f, tag)
    status = f.wait()
    if status:
        logging.warning("git log exited with status %d", status)


def gen_update_branch_changes(oldrev, newrev, refname, branch):
    # A branch has been updated. If it was a fast-forward update,
    # generate Change events for everything between oldrev and newrev.
    #
    # In case of a forced update, first generate a "fake" Change event
    # rewinding the branch to the common ancestor of oldrev and
    # newrev. Then, generate Change events for each commit between the
    # common ancestor and newrev.

    logging.info("Branch `%s' updated %s .. %s", branch, oldrev[:8], newrev[:8])

    mergebasecommand = subprocess.Popen(
        ["git", "merge-base", oldrev, newrev], stdout=subprocess.PIPE
    )
    (baserev, err) = mergebasecommand.communicate()
    baserev = baserev.strip()  # remove newline
    baserev = baserev.decode(encoding)

    logging.debug("oldrev=%s newrev=%s baserev=%s", oldrev, newrev, baserev)
    if baserev != oldrev:
        c = {
            'revision': baserev,
            'comments': "Rewind branch",
            'branch': text_type(branch),
            'who': "dummy",
        }
        logging.info("Branch %s was rewound to %s", branch, baserev[:8])
        files = []
        f = subprocess.Popen(
            shlex.split(f"git diff --raw {oldrev}..{baserev}"),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        while True:
            line = f.stdout.readline().decode(encoding)
            if not line:
                break

            file = re.match(r"^:.*[MAD]\s+(.+)$", line).group(1)
            logging.debug("  Rewound file: %s", file)
            files.append(text_type(file))

        status = f.wait()
        if status:
            logging.warning("git diff exited with status %d", status)

        if category:
            c['category'] = text_type(category)

        if repository:
            c['repository'] = text_type(repository)

        if project:
            c['project'] = text_type(project)

        if codebase:
            c['codebase'] = text_type(codebase)

        if files:
            c['files'] = files
            changes.append(c)

    if newrev != baserev:
        # Not a pure rewind
        options = "--reverse --pretty=oneline"
        if first_parent:
            # Add the --first-parent to avoid adding the merge commits which
            # have already been tested.
            options += ' --first-parent'
        f = subprocess.Popen(
            shlex.split(f"git rev-list {options} {baserev}..{newrev}"),
            stdout=subprocess.PIPE,
        )
        gen_changes(f, branch)

        status = f.wait()
        if status:
            logging.warning("git rev-list exited with status %d", status)


def cleanup(res):
    reactor.stop()


def process_branch_change(oldrev, newrev, refname, branch):
    # Find out if the branch was created, deleted or updated.
    if re.match(r"^0*$", newrev):
        logging.info("Branch `%s' deleted, ignoring", branch)
    elif re.match(r"^0*$", oldrev):
        gen_create_branch_changes(newrev, refname, branch)
    else:
        gen_update_branch_changes(oldrev, newrev, refname, branch)


def process_tag_change(oldrev, newrev, refname, tag):
    # Process a new tag, or ignore a deleted tag
    if re.match(r"^0*$", newrev):
        logging.info("Tag `%s' deleted, ignoring", tag)
    elif re.match(r"^0*$", oldrev):
        gen_create_tag_changes(newrev, refname, tag)


def process_change(oldrev, newrev, refname):
    # Identify the change as a branch, tag or other, and process it
    m = re.match(r"^refs/(heads|tags)/(.+)$", refname)
    if not m:
        logging.info("Ignoring refname `%s': Not a branch or tag", refname)
        return

    if m.group(1) == 'heads':
        branch = m.group(2)
        process_branch_change(oldrev, newrev, refname, branch)
    elif m.group(1) == 'tags':
        tag = m.group(2)
        process_tag_change(oldrev, newrev, refname, tag)


def process_changes():
    # Read branch updates from stdin and generate Change events
    while True:
        line = sys.stdin.readline()
        line = line.rstrip()
        if not line:
            break

        args = line.split(None, 2)
        [oldrev, newrev, refname] = args
        process_change(oldrev, newrev, refname)


def send_changes():
    # Submit the changes, if any
    if not changes:
        logging.info("No changes found")
        return

    host, port = master.split(':')
    port = int(port)

    f = pb.PBClientFactory()
    d = f.login(credentials.UsernamePassword(username.encode('utf-8'), auth.encode('utf-8')))
    reactor.connectTCP(host, port, f)

    d.addErrback(connectFailed)
    d.addCallback(connected)
    d.addBoth(cleanup)

    reactor.run()


def parse_options():
    parser = OptionParser()
    parser.add_option(
        "-l", "--logfile", action="store", type="string", help="Log to the specified file"
    )
    parser.add_option(
        "-v", "--verbose", action="count", help="Be more verbose. Ignored if -l is not specified."
    )
    master_help = f"Build master to push to. Default is {master}"
    parser.add_option("-m", "--master", action="store", type="string", help=master_help)
    parser.add_option(
        "-c", "--category", action="store", type="string", help="Scheduler category to notify."
    )
    parser.add_option(
        "-r", "--repository", action="store", type="string", help="Git repository URL to send."
    )
    parser.add_option("-p", "--project", action="store", type="string", help="Project to send.")
    parser.add_option("--codebase", action="store", type="string", help="Codebase to send.")
    encoding_help = f"Encoding to use when converting strings to unicode. Default is {encoding}."
    parser.add_option("-e", "--encoding", action="store", type="string", help=encoding_help)
    username_help = f"Username used in PB connection auth, defaults to {username}."
    parser.add_option("-u", "--username", action="store", type="string", help=username_help)
    auth_help = f"Password used in PB connection auth, defaults to {auth}."
    # 'a' instead of 'p' due to collisions with the project short option
    parser.add_option("-a", "--auth", action="store", type="string", help=auth_help)
    first_parent_help = "If set, don't trigger builds for merged in commits"
    parser.add_option("--first-parent", action="store_true", help=first_parent_help)
    options, args = parser.parse_args()
    return options


# Log errors and critical messages to stderr. Optionally log
# information to a file as well (we'll set that up later.)
stderr = logging.StreamHandler(sys.stderr)
fmt = logging.Formatter("git_buildbot: %(levelname)s: %(message)s")
stderr.setLevel(logging.WARNING)
stderr.setFormatter(fmt)
logging.getLogger().addHandler(stderr)
logging.getLogger().setLevel(logging.NOTSET)

try:
    options = parse_options()
    level = logging.WARNING
    if options.verbose:
        level -= 10 * options.verbose
        level = max(level, 0)

    if options.logfile:
        logfile = logging.FileHandler(options.logfile)
        logfile.setLevel(level)
        fmt = logging.Formatter("%(asctime)s %(levelname)s: %(message)s")
        logfile.setFormatter(fmt)
        logging.getLogger().addHandler(logfile)

    if options.master:
        master = options.master

    if options.category:
        category = options.category

    if options.repository:
        repository = options.repository

    if options.project:
        project = options.project

    if options.codebase:
        codebase = options.codebase

    if options.username:
        username = options.username

    if options.auth:
        auth = options.auth

    if options.encoding:
        encoding = options.encoding

    if options.first_parent:
        first_parent = options.first_parent

    process_changes()
    send_changes()
except Exception:
    logging.exception("Unhandled exception")
    sys.exit(1)



from buildbot.www.authz.authz import Authz
from buildbot.www.authz.authz import Forbidden
from buildbot.www.authz.authz import fnmatchStrMatcher
from buildbot.www.authz.authz import reStrMatcher

__all__ = ["Authz", "Forbidden", "fnmatchStrMatcher", "reStrMatcher"]



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class TestName(Row):
    table = 'test_names'

    id_column = 'id'

    def __init__(self, id=None, builderid=None, name='nam'):
        super().__init__(id=id, builderid=builderid, name=name)


class TestCodePath(Row):
    table = 'test_code_paths'

    id_column = 'id'

    def __init__(self, id=None, builderid=None, path='path/to/file'):
        super().__init__(id=id, builderid=builderid, path=path)


class TestResult(Row):
    table = 'test_results'

    id_column = 'id'

    def __init__(
        self,
        id=None,
        builderid=None,
        test_result_setid=None,
        test_nameid=None,
        test_code_pathid=None,
        line=None,
        duration_ns=None,
        value=None,
    ):
        super().__init__(
            id=id,
            builderid=builderid,
            test_result_setid=test_result_setid,
            test_nameid=test_nameid,
            test_code_pathid=test_code_pathid,
            line=line,
            duration_ns=duration_ns,
            value=value,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.config import BuilderConfig
from buildbot.data import resultspec
from buildbot.process.factory import BuildFactory
from buildbot.process.results import SUCCESS
from buildbot.process.workerforbuilder import PingException
from buildbot.schedulers import triggerable
from buildbot.steps import trigger
from buildbot.test.fake.worker import WorkerController
from buildbot.test.util.integration import RunFakeMasterTestCase
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import Coroutine
    from typing import TypeVar

    from typing_extensions import ParamSpec

    _T = TypeVar('_T')
    _P = ParamSpec('_P')


class Tests(RunFakeMasterTestCase):
    @defer.inlineCallbacks
    def do_terminates_ping_on_shutdown(self, quick_mode):
        """
        During shutdown we want to terminate any outstanding pings.
        """
        controller = WorkerController(self, 'local')

        config_dict = {
            'builders': [
                BuilderConfig(name="testy", workernames=['local'], factory=BuildFactory()),
            ],
            'workers': [controller.worker],
            'protocols': {'null': {}},
            'multiMaster': True,
        }
        yield self.setup_master(config_dict)
        builder_id = yield self.master.data.updates.findBuilderId('testy')

        yield controller.connect_worker()
        controller.sever_connection()
        yield self.create_build_request([builder_id])

        # give time for any delayed actions to complete
        self.reactor.advance(1)

        yield self.master.botmaster.cleanShutdown(quickMode=quick_mode, stopReactor=False)
        self.flushLoggedErrors(PingException)

    def test_terminates_ping_on_shutdown_quick_mode(self):
        return self.do_terminates_ping_on_shutdown(quick_mode=True)

    def test_terminates_ping_on_shutdown_slow_mode(self):
        return self.do_terminates_ping_on_shutdown(quick_mode=False)

    def _wait_step(self, wait_step: float = 0.1, timeout_seconds: float = 5.0):
        for _ in range(0, int(timeout_seconds * 1000), int(wait_step * 1000)):
            self.reactor.advance(wait_step)
            yield

    async def _query_until_result(
        self,
        fn: Callable[_P, Coroutine[Any, Any, _T]],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _T:
        for _ in self._wait_step():
            result = await fn(*args, **kwargs)
            if result:
                return result
        self.fail('Fail to get result in appropriate timeout')

    @async_to_deferred
    async def test_shutdown_busy_with_child(self) -> None:
        """
        Test that clean shutdown complete correctly
        even when a running Build trigger another
        and wait for it's completion
        """

        parent_controller = WorkerController(self, 'parent_worker')
        child_controller = WorkerController(self, 'child_worker')

        config_dict = {
            'builders': [
                BuilderConfig(
                    name="parent",
                    workernames=[parent_controller.worker.name],
                    factory=BuildFactory([
                        trigger.Trigger(schedulerNames=['triggerable'], waitForFinish=True)
                    ]),
                ),
                BuilderConfig(
                    name="child", workernames=[child_controller.worker.name], factory=BuildFactory()
                ),
            ],
            'workers': [parent_controller.worker, child_controller.worker],
            'schedulers': [triggerable.Triggerable(name='triggerable', builderNames=['child'])],
            'protocols': {'null': {}},
            'multiMaster': True,
            'collapseRequests': False,
        }
        await self.setup_master(config_dict)

        parent_builder_id = await self.master.data.updates.findBuilderId('parent')
        child_builder_id = await self.master.data.updates.findBuilderId('child')

        await parent_controller.connect_worker()
        # Pause worker of Child builder so we know the build won't start before we start shutdown
        await child_controller.disconnect_worker()

        # Create a Child build without Parent so we can later make sure it was not executed
        _, first_child_brids = await self.create_build_request([child_builder_id])
        self.assertEqual(len(first_child_brids), 1)

        _, _parent_brids = await self.create_build_request([parent_builder_id])
        self.assertEqual(len(_parent_brids), 1)
        parent_brid = _parent_brids[parent_builder_id]

        # wait until Parent trigger it's Child build
        parent_buildid = (
            await self._query_until_result(
                self.master.data.get,
                ("builds",),
                filters=[resultspec.Filter('buildrequestid', 'eq', [parent_brid])],
            )
        )[0]['buildid']

        # now get the child_buildset
        child_buildsetid = (
            await self._query_until_result(
                self.master.data.get,
                ("buildsets",),
                filters=[resultspec.Filter('parent_buildid', 'eq', [parent_buildid])],
            )
        )[0]['bsid']

        # and finally, the child BuildReques
        child_buildrequest = (
            await self._query_until_result(
                self.master.data.get,
                ("buildrequests",),
                filters=[resultspec.Filter('buildsetid', 'eq', [child_buildsetid])],
            )
        )[0]

        # now we know the Parent's Child BuildRequest exists,
        # create a second Child without Parent for good measure
        _, second_child_brids = await self.create_build_request([child_builder_id])
        self.assertEqual(len(second_child_brids), 1)

        # Now start the clean shutdown
        shutdown_deferred: defer.Deferred[None] = self.master.botmaster.cleanShutdown(
            quickMode=False,
            stopReactor=False,
        )

        # Connect back Child worker so the build can happen
        await child_controller.connect_worker()

        # wait for the child request to be claimed, and completed
        for _ in self._wait_step():
            if child_buildrequest['claimed'] and child_buildrequest['complete']:
                break
            child_buildrequest = await self.master.data.get(
                ("buildrequests", child_buildrequest['buildrequestid']),
            )
            self.assertIsNotNone(child_buildrequest)
        self.assertEqual(child_buildrequest['results'], SUCCESS)

        # make sure parent-less BuildRequest weren't built
        first_child_request = await self.master.data.get(
            ("buildrequests", first_child_brids[child_builder_id]),
        )
        self.assertIsNotNone(first_child_request)
        self.assertFalse(first_child_request['claimed'])
        self.assertFalse(first_child_request['complete'])

        second_child_request = await self.master.data.get(
            ("buildrequests", second_child_brids[child_builder_id]),
        )
        self.assertIsNotNone(second_child_request)
        self.assertFalse(second_child_request['claimed'])
        self.assertFalse(second_child_request['complete'])

        # confirm Master shutdown
        await shutdown_deferred
        self.assertTrue(shutdown_deferred.called)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import datetime
import locale
import os
import sys
from unittest import mock

from twisted.internet import reactor
from twisted.internet import task
from twisted.trial import unittest

from buildbot import util


class formatInterval(unittest.TestCase):
    def test_zero(self) -> None:
        self.assertEqual(util.formatInterval(0), "0 secs")

    def test_seconds_singular(self) -> None:
        self.assertEqual(util.formatInterval(1), "1 secs")

    def test_seconds(self) -> None:
        self.assertEqual(util.formatInterval(7), "7 secs")

    def test_minutes_one(self) -> None:
        self.assertEqual(util.formatInterval(60), "60 secs")

    def test_minutes_over_one(self) -> None:
        self.assertEqual(util.formatInterval(61), "1 mins, 1 secs")

    def test_minutes(self) -> None:
        self.assertEqual(util.formatInterval(300), "5 mins, 0 secs")

    def test_hours_one(self) -> None:
        self.assertEqual(util.formatInterval(3600), "60 mins, 0 secs")

    def test_hours_over_one_sec(self) -> None:
        self.assertEqual(util.formatInterval(3601), "1 hrs, 1 secs")

    def test_hours_over_one_min(self) -> None:
        self.assertEqual(util.formatInterval(3660), "1 hrs, 60 secs")

    def test_hours(self) -> None:
        self.assertEqual(util.formatInterval(7200), "2 hrs, 0 secs")

    def test_mixed(self) -> None:
        self.assertEqual(util.formatInterval(7392), "2 hrs, 3 mins, 12 secs")


class TestHumanReadableDelta(unittest.TestCase):
    def test_timeDeltaToHumanReadable(self) -> None:
        """
        It will return a human readable time difference.
        """
        result = util.human_readable_delta(1, 1)
        self.assertEqual('super fast', result)

        result = util.human_readable_delta(1, 2)
        self.assertEqual('1 seconds', result)

        result = util.human_readable_delta(1, 61)
        self.assertEqual('1 minutes', result)

        result = util.human_readable_delta(1, 62)
        self.assertEqual('1 minutes, 1 seconds', result)

        result = util.human_readable_delta(1, 60 * 60 + 1)
        self.assertEqual('1 hours', result)

        result = util.human_readable_delta(1, 60 * 60 + 61)
        self.assertEqual('1 hours, 1 minutes', result)

        result = util.human_readable_delta(1, 60 * 60 + 62)
        self.assertEqual('1 hours, 1 minutes, 1 seconds', result)

        result = util.human_readable_delta(1, 24 * 60 * 60 + 1)
        self.assertEqual('1 days', result)

        result = util.human_readable_delta(1, 24 * 60 * 60 + 2)
        self.assertEqual('1 days, 1 seconds', result)


class TestFuzzyInterval(unittest.TestCase):
    def test_moment(self) -> None:
        self.assertEqual(util.fuzzyInterval(1), "a moment")

    def test_seconds(self) -> None:
        self.assertEqual(util.fuzzyInterval(17), "17 seconds")

    def test_seconds_rounded(self) -> None:
        self.assertEqual(util.fuzzyInterval(48), "50 seconds")

    def test_minute(self) -> None:
        self.assertEqual(util.fuzzyInterval(58), "a minute")

    def test_minutes(self) -> None:
        self.assertEqual(util.fuzzyInterval(3 * 60 + 24), "3 minutes")

    def test_minutes_rounded(self) -> None:
        self.assertEqual(util.fuzzyInterval(32 * 60 + 24), "30 minutes")

    def test_hour(self) -> None:
        self.assertEqual(util.fuzzyInterval(3600 + 1200), "an hour")

    def test_hours(self) -> None:
        self.assertEqual(util.fuzzyInterval(9 * 3600 - 720), "9 hours")

    def test_day(self) -> None:
        self.assertEqual(util.fuzzyInterval(32 * 3600 + 124), "a day")

    def test_days(self) -> None:
        self.assertEqual(util.fuzzyInterval((19 + 24) * 3600 + 124), "2 days")

    def test_month(self) -> None:
        self.assertEqual(util.fuzzyInterval(36 * 24 * 3600 + 124), "a month")

    def test_months(self) -> None:
        self.assertEqual(util.fuzzyInterval(86 * 24 * 3600 + 124), "3 months")

    def test_year(self) -> None:
        self.assertEqual(util.fuzzyInterval(370 * 24 * 3600), "a year")

    def test_years(self) -> None:
        self.assertEqual(util.fuzzyInterval((2 * 365 + 96) * 24 * 3600), "2 years")


class safeTranslate(unittest.TestCase):
    def test_str_good(self) -> None:
        self.assertEqual(util.safeTranslate("full"), b"full")

    def test_str_bad(self) -> None:
        self.assertEqual(util.safeTranslate("speed=slow;quality=high"), b"speed_slow_quality_high")

    def test_str_pathological(self) -> None:
        # if you needed proof this wasn't for use with sensitive data
        self.assertEqual(
            util.safeTranslate("p\ath\x01ogy"), b"p\ath\x01ogy"
        )  # bad chars still here!

    def test_unicode_good(self) -> None:
        self.assertEqual(util.safeTranslate("full"), b"full")

    def test_unicode_bad(self) -> None:
        self.assertEqual(util.safeTranslate("speed=slow;quality=high"), b"speed_slow_quality_high")

    def test_unicode_pathological(self) -> None:
        self.assertEqual(util.safeTranslate("\u0109"), b"\xc4\x89")  # yuck!


class naturalSort(unittest.TestCase):
    def test_alpha(self) -> None:
        self.assertEqual(util.naturalSort(['x', 'aa', 'ab']), ['aa', 'ab', 'x'])

    def test_numeric(self) -> None:
        self.assertEqual(
            util.naturalSort(['1', '10', '11', '2', '20']), ['1', '2', '10', '11', '20']
        )

    def test_alphanum(self) -> None:
        l1 = 'aa10ab aa1ab aa10aa f a aa3 aa30 aa3a aa30a'.split()
        l2 = 'a aa1ab aa3 aa3a aa10aa aa10ab aa30 aa30a f'.split()
        self.assertEqual(util.naturalSort(l1), l2)


class none_or_str(unittest.TestCase):
    def test_none(self) -> None:
        self.assertEqual(util.none_or_str(None), None)

    def test_str(self) -> None:
        self.assertEqual(util.none_or_str("hi"), "hi")

    def test_int(self) -> None:
        self.assertEqual(util.none_or_str(199), "199")


class TimeFunctions(unittest.TestCase):
    def test_UTC(self) -> None:
        self.assertEqual(util.UTC.utcoffset(datetime.datetime.now()), datetime.timedelta(0))
        self.assertEqual(util.UTC.dst(datetime.datetime.now()), datetime.timedelta(0))
        self.assertEqual(util.UTC.tzname(datetime.datetime.now(datetime.timezone.utc)), "UTC")

    def test_epoch2datetime(self) -> None:
        self.assertEqual(
            util.epoch2datetime(0), datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=util.UTC)
        )
        self.assertEqual(
            util.epoch2datetime(1300000000),
            datetime.datetime(2011, 3, 13, 7, 6, 40, tzinfo=util.UTC),
        )

    def test_datetime2epoch(self) -> None:
        dt = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=util.UTC)
        self.assertEqual(util.datetime2epoch(dt), 0)
        dt = datetime.datetime(2011, 3, 13, 7, 6, 40, tzinfo=util.UTC)
        self.assertEqual(util.datetime2epoch(dt), 1300000000)


class DiffSets(unittest.TestCase):
    def test_empty(self) -> None:
        removed, added = util.diffSets(set([]), set([]))
        self.assertEqual((removed, added), (set([]), set([])))

    def test_no_lists(self) -> None:
        removed, added = util.diffSets([1, 2], [2, 3])
        self.assertEqual((removed, added), (set([1]), set([3])))

    def test_no_overlap(self) -> None:
        removed, added = util.diffSets(set([1, 2]), set([3, 4]))
        self.assertEqual((removed, added), (set([1, 2]), set([3, 4])))

    def test_no_change(self) -> None:
        removed, added = util.diffSets(set([1, 2]), set([1, 2]))
        self.assertEqual((removed, added), (set([]), set([])))

    def test_added(self) -> None:
        removed, added = util.diffSets(set([1, 2]), set([1, 2, 3]))
        self.assertEqual((removed, added), (set([]), set([3])))

    def test_removed(self) -> None:
        removed, added = util.diffSets(set([1, 2]), set([1]))
        self.assertEqual((removed, added), (set([2]), set([])))


class MakeList(unittest.TestCase):
    def test_empty_string(self) -> None:
        self.assertEqual(util.makeList(''), [''])

    def test_None(self) -> None:
        self.assertEqual(util.makeList(None), [])

    def test_string(self) -> None:
        self.assertEqual(util.makeList('hello'), ['hello'])

    def test_unicode(self) -> None:
        self.assertEqual(util.makeList('\N{SNOWMAN}'), ['\N{SNOWMAN}'])

    def test_list(self) -> None:
        self.assertEqual(util.makeList(['a', 'b']), ['a', 'b'])

    def test_tuple(self) -> None:
        self.assertEqual(util.makeList(('a', 'b')), ['a', 'b'])

    def test_copy(self) -> None:
        input = ['a', 'b']
        output = util.makeList(input)
        input.append('c')
        self.assertEqual(output, ['a', 'b'])


class Flatten(unittest.TestCase):
    def test_simple(self) -> None:
        self.assertEqual(util.flatten([1, 2, 3]), [1, 2, 3])

    def test_deep(self) -> None:
        self.assertEqual(util.flatten([[1, 2], 3, [[4]]]), [1, 2, 3, 4])

    # def test_deeply_nested(self) -> None:
    #     self.assertEqual(util.flatten([5, [6, (7, 8)]]),
    #                      [5, 6, 7, 8])

    # def test_tuples(self) -> None:
    #     self.assertEqual(util.flatten([(1, 2), 3]), [1, 2, 3])

    def test_dict(self) -> None:
        d = {'a': [5, 6, 7], 'b': [7, 8, 9]}
        self.assertEqual(util.flatten(d), d)

    def test_string(self) -> None:
        self.assertEqual(util.flatten("abc"), "abc")


class Ascii2Unicode(unittest.TestCase):
    def test_unicode(self) -> None:
        rv = util.bytes2unicode('\N{SNOWMAN}', encoding='ascii')
        self.assertEqual((rv, type(rv)), ('\N{SNOWMAN}', str))

    def test_ascii(self) -> None:
        rv = util.bytes2unicode('abcd', encoding='ascii')
        self.assertEqual((rv, type(rv)), ('abcd', str))

    def test_nonascii(self) -> None:
        with self.assertRaises(UnicodeDecodeError):
            util.bytes2unicode(b'a\x85', encoding='ascii')

    def test_None(self) -> None:
        self.assertEqual(util.bytes2unicode(None, encoding='ascii'), None)

    def test_bytes2unicode(self) -> None:
        rv1 = util.bytes2unicode(b'abcd')
        rv2 = util.bytes2unicode('efgh')

        self.assertEqual(type(rv1), str)
        self.assertEqual(type(rv2), str)


class StringToBoolean(unittest.TestCase):
    def test_it(self) -> None:
        stringValues = [
            (b'on', True),
            (b'true', True),
            (b'yes', True),
            (b'1', True),
            (b'off', False),
            (b'false', False),
            (b'no', False),
            (b'0', False),
            (b'ON', True),
            (b'TRUE', True),
            (b'YES', True),
            (b'OFF', False),
            (b'FALSE', False),
            (b'NO', False),
        ]
        for s, b in stringValues:
            self.assertEqual(util.string2boolean(s), b, repr(s))

    def test_ascii(self) -> None:
        rv = util.bytes2unicode(b'abcd', encoding='ascii')
        self.assertEqual((rv, type(rv)), ('abcd', str))

    def test_nonascii(self) -> None:
        with self.assertRaises(UnicodeDecodeError):
            util.bytes2unicode(b'a\x85', encoding='ascii')

    def test_None(self) -> None:
        self.assertEqual(util.bytes2unicode(None, encoding='ascii'), None)


class AsyncSleep(unittest.TestCase):
    def test_sleep(self) -> None:
        clock = task.Clock()
        self.patch(reactor, 'callLater', clock.callLater)
        d = util.asyncSleep(2)
        self.assertFalse(d.called)
        clock.advance(1)
        self.assertFalse(d.called)
        clock.advance(1)
        self.assertTrue(d.called)


class FunctionalEnvironment(unittest.TestCase):
    def test_working_locale(self) -> None:
        environ = {'LANG': 'en_GB.UTF-8'}
        self.patch(os, 'environ', environ)
        config = mock.Mock()
        util.check_functional_environment(config)
        self.assertEqual(config.error.called, False)

    def test_broken_locale(self) -> None:
        def err() -> None:
            raise KeyError

        if sys.version_info >= (3, 11, 0):
            self.patch(locale, 'getencoding', err)
        else:
            self.patch(locale, 'getdefaultlocale', err)
        config = mock.Mock()
        util.check_functional_environment(config)
        config.error.assert_called_with(mock.ANY)


class StripUrlPassword(unittest.TestCase):
    def test_simple_url(self) -> None:
        self.assertEqual(util.stripUrlPassword('http://foo.com/bar'), 'http://foo.com/bar')

    def test_username(self) -> None:
        self.assertEqual(util.stripUrlPassword('http://d@foo.com/bar'), 'http://d@foo.com/bar')

    def test_username_with_at(self) -> None:
        self.assertEqual(
            util.stripUrlPassword('http://d@bb.net@foo.com/bar'), 'http://d@bb.net@foo.com/bar'
        )

    def test_username_pass(self) -> None:
        self.assertEqual(
            util.stripUrlPassword('http://d:secret@foo.com/bar'), 'http://d:xxxx@foo.com/bar'
        )

    def test_username_pass_with_at(self) -> None:
        self.assertEqual(
            util.stripUrlPassword('http://d@bb.net:scrt@foo.com/bar'),
            'http://d@bb.net:xxxx@foo.com/bar',
        )


class JoinList(unittest.TestCase):
    def test_list(self) -> None:
        self.assertEqual(util.join_list(['aa', 'bb']), 'aa bb')

    def test_tuple(self) -> None:
        self.assertEqual(util.join_list(('aa', 'bb')), 'aa bb')

    def test_string(self) -> None:
        self.assertEqual(util.join_list('abc'), 'abc')

    def test_unicode(self) -> None:
        self.assertEqual(util.join_list('abc'), 'abc')

    def test_nonascii(self) -> None:
        with self.assertRaises(UnicodeDecodeError):
            util.join_list([b'\xff'])


class CommandToString(unittest.TestCase):
    def test_short_string(self) -> None:
        self.assertEqual(util.command_to_string("ab cd"), "'ab cd'")

    def test_long_string(self) -> None:
        self.assertEqual(util.command_to_string("ab cd ef"), "'ab cd ...'")

    def test_list(self) -> None:
        self.assertEqual(util.command_to_string(['ab', 'cd', 'ef']), "'ab cd ...'")

    def test_nested_list(self) -> None:
        self.assertEqual(util.command_to_string(['ab', ['cd', ['ef']]]), "'ab cd ...'")

    def test_object(self) -> None:
        # this looks like a renderable
        self.assertEqual(util.command_to_string(object()), None)

    def test_list_with_objects(self) -> None:
        self.assertRegex(util.command_to_string(['ab', object(), 'cd']), r"'ab <object .*> \.\.\.'")

    def test_invalid_ascii(self) -> None:
        self.assertEqual(util.command_to_string(b'a\xffc'), "'a\ufffdc'")


class TestRewrap(unittest.TestCase):
    def test_main(self) -> None:
        tests = [
            ("", "", None),
            ("\n", "\n", None),
            ("\n  ", "\n", None),
            ("  \n", "\n", None),
            ("  \n  ", "\n", None),
            (
                """
                multiline
                with
                indent
                """,
                "\nmultiline with indent",
                None,
            ),
            (
                """\
                multiline
                with
                indent

                """,
                "multiline with indent\n",
                None,
            ),
            (
                """\
                 multiline
                 with
                 indent

                 """,
                "multiline with indent\n",
                None,
            ),
            (
                """\
                multiline
                with
                indent

                  and
                   formatting
                """,
                "multiline with indent\n  and\n   formatting\n",
                None,
            ),
            (
                """\
                multiline
                with
                indent
                and wrapping

                  and
                   formatting
                """,
                "multiline with\nindent and\nwrapping\n  and\n   formatting\n",
                15,
            ),
        ]

        for text, expected, width in tests:
            self.assertEqual(util.rewrap(text, width=width), expected)


class TestMerge(unittest.TestCase):
    def test_merge(self) -> None:
        self.assertEqual(
            util.dictionary_merge({'a': {'b': 1}}, {'a': {'c': 2}}), {'a': {'b': 1, 'c': 2}}
        )

    def test_overwrite(self) -> None:
        self.assertEqual(util.dictionary_merge({'a': {'b': 1}}, {'a': 1}), {'a': 1})

    def test_overwrite2(self) -> None:
        self.assertEqual(
            util.dictionary_merge({'a': {'b': 1, 'c': 2}}, {'a': {'b': [1, 2, 3]}}),
            {'a': {'b': [1, 2, 3], 'c': 2}},
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os


class FakePrivateTemporaryDirectory:
    def __init__(self, suffix=None, prefix=None, dir=None, mode=0o700):
        dir = dir or '/'
        prefix = prefix or ''
        suffix = suffix or ''

        self.name = os.path.join(dir, prefix + '@@@' + suffix)
        self.mode = mode

    def __enter__(self):
        return self.name

    def __exit__(self, exc, value, tb):
        pass

    def cleanup(self):
        pass


class MockPrivateTemporaryDirectory:
    def __init__(self):
        self.dirs = []

    def __call__(self, *args, **kwargs):
        ret = FakePrivateTemporaryDirectory(*args, **kwargs)
        self.dirs.append((ret.name, ret.mode))
        return ret



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.python import deprecate
from twisted.python import versions

from buildbot.db import base
from buildbot.util.twisted import async_to_deferred
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    from twisted.internet import defer


@dataclass
class TestResultSetModel:
    id: int
    builderid: int
    buildid: int
    stepid: int
    description: str | None
    category: str
    value_unit: str
    tests_passed: int | None
    tests_failed: int | None
    complete: bool = False

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'TestResultSetsConnectorComponent '
                'getTestResultSet, and getTestResultSets '
                'no longer return TestResultSet as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


@deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), TestResultSetModel)
class TestResultSetDict(TestResultSetModel):
    pass


class TestResultSetAlreadyCompleted(Exception):
    pass


class TestResultSetsConnectorComponent(base.DBConnectorComponent):
    def addTestResultSet(
        self, builderid, buildid, stepid, description, category, value_unit
    ) -> defer.Deferred[int]:
        # Returns the id of the new result set
        def thd(conn) -> int:
            sets_table = self.db.model.test_result_sets

            insert_values = {
                'builderid': builderid,
                'buildid': buildid,
                'stepid': stepid,
                'description': description,
                'category': category,
                'value_unit': value_unit,
                'complete': 0,
            }

            q = sets_table.insert().values(insert_values)
            r = conn.execute(q)
            conn.commit()
            return r.inserted_primary_key[0]

        return self.db.pool.do(thd)

    def getTestResultSet(self, test_result_setid: int) -> defer.Deferred[TestResultSetModel | None]:
        def thd(conn) -> TestResultSetModel | None:
            sets_table = self.db.model.test_result_sets
            q = sets_table.select().where(sets_table.c.id == test_result_setid)
            res = conn.execute(q)
            row = res.fetchone()
            if not row:
                return None
            return self._model_from_row(row)

        return self.db.pool.do(thd)

    def getTestResultSets(
        self,
        builderid: int | None = None,
        buildid: int | None = None,
        stepid: int | None = None,
        complete: bool | None = None,
        result_spec=None,
    ) -> defer.Deferred[list[TestResultSetModel]]:
        def thd(conn) -> list[TestResultSetModel]:
            sets_table = self.db.model.test_result_sets
            q = sets_table.select()
            if builderid is not None:
                q = q.where(sets_table.c.builderid == builderid)
            if buildid is not None:
                q = q.where(sets_table.c.buildid == buildid)
            if stepid is not None:
                q = q.where(sets_table.c.stepid == stepid)
            if complete is not None:
                q = q.where(sets_table.c.complete == (1 if complete else 0))
            if result_spec is not None:
                return result_spec.thd_execute(conn, q, self._model_from_row)
            res = conn.execute(q)
            return [self._model_from_row(row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    @async_to_deferred
    async def get_test_result_sets_for_commits(
        self, *, commit_ids: list[int]
    ) -> list[TestResultSetModel]:
        def thd(conn) -> list[TestResultSetModel]:
            # FIXME: the code below currently is not sufficiently robust because it relies on
            # revisions being sufficiently random so that they do not repeat across whole Buildbot
            # database. At least the following would be needed to resolve the ambiguities
            #  - attach sourcestamp to a codebase through a codebase ID
            #  - attach sourcestamp to a commit through a commit ID
            j = self.db.model.codebase_commits
            j = j.join(
                self.db.model.sourcestamps,
                self.db.model.codebase_commits.c.revision == self.db.model.sourcestamps.c.revision,
            )
            j = j.join(self.db.model.buildset_sourcestamps)
            j = j.join(self.db.model.buildsets)
            j = j.join(self.db.model.buildrequests)
            j = j.join(self.db.model.builds)
            j = j.join(self.db.model.test_result_sets)
            q = (
                sa.select(self.db.model.test_result_sets)
                .select_from(j)
                .where(self.db.model.codebase_commits.c.id.in_(commit_ids))
            )

            res = conn.execute(q)
            return [self._model_from_row(row) for row in res.fetchall()]

        return await self.db.pool.do(thd)

    def completeTestResultSet(
        self, test_result_setid, tests_passed=None, tests_failed=None
    ) -> defer.Deferred[None]:
        def thd(conn) -> None:
            sets_table = self.db.model.test_result_sets

            values = {'complete': 1}
            if tests_passed is not None:
                values['tests_passed'] = tests_passed
            if tests_failed is not None:
                values['tests_failed'] = tests_failed

            q = sets_table.update().values(values)
            q = q.where((sets_table.c.id == test_result_setid) & (sets_table.c.complete == 0))

            res = conn.execute(q)
            conn.commit()
            if res.rowcount == 0:
                raise TestResultSetAlreadyCompleted(
                    f'Test result set {test_result_setid} is already completed or does not exist'
                )

        return self.db.pool.do(thd)

    def _model_from_row(self, row):
        return TestResultSetModel(
            id=row.id,
            builderid=row.builderid,
            buildid=row.buildid,
            stepid=row.stepid,
            description=row.description,
            category=row.category,
            value_unit=row.value_unit,
            tests_passed=row.tests_passed,
            tests_failed=row.tests_failed,
            complete=bool(row.complete),
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class Step(Row):
    table = "steps"

    id_column = 'id'

    def __init__(
        self,
        id=None,
        number=29,
        name='step29',
        buildid=None,
        started_at=1304262222,
        locks_acquired_at=None,
        complete_at=None,
        state_string='',
        results=None,
        urls_json='[]',
        hidden=0,
    ):
        super().__init__(
            id=id,
            number=number,
            name=name,
            buildid=buildid,
            started_at=started_at,
            locks_acquired_at=locks_acquired_at,
            complete_at=complete_at,
            state_string=state_string,
            results=results,
            urls_json=urls_json,
            hidden=hidden,
        )



# This program is free software; you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members
# Portions Copyright Marius Rieder <marius.rieder@durchmesser.ch>
"""
Steps and objects related to pbuilder
"""

from __future__ import annotations

import re
import stat
import time

from twisted.internet import defer
from twisted.python import log

from buildbot import config
from buildbot.process import logobserver
from buildbot.process import remotecommand
from buildbot.process import results
from buildbot.steps.shell import WarningCountingShellCommand


class DebPbuilder(WarningCountingShellCommand):
    """Build a debian package with pbuilder inside of a chroot."""

    name = "pbuilder"

    haltOnFailure = True
    flunkOnFailure = True
    description = ["building"]
    descriptionDone = ["built"]

    warningPattern = r".*(warning[: ]|\sW: ).*"

    architecture = None
    distribution: str | None = 'stable'
    basetgz = None
    _default_basetgz = "/var/cache/pbuilder/{distribution}-{architecture}-buildbot.tgz"
    mirror = "http://cdn.debian.net/debian/"
    othermirror = ""
    extrapackages: list[str] = []
    keyring = None
    components: str | None = None

    maxAge = 60 * 60 * 24 * 7
    pbuilder = '/usr/sbin/pbuilder'
    baseOption = '--basetgz'

    renderables = [
        'architecture',
        'distribution',
        'basetgz',
        'mirror',
        'othermirror',
        'extrapackages',
        'keyring',
        'components',
    ]

    def __init__(
        self,
        architecture=None,
        distribution=None,
        basetgz=None,
        mirror=None,
        othermirror=None,
        extrapackages=None,
        keyring=None,
        components=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if architecture:
            self.architecture = architecture
        if distribution:
            self.distribution = distribution
        if mirror:
            self.mirror = mirror
        if othermirror:
            self.othermirror = "|".join(othermirror)
        if extrapackages:
            self.extrapackages = extrapackages
        if keyring:
            self.keyring = keyring
        if components:
            self.components = components
        if basetgz:
            self.basetgz = basetgz

        if not self.distribution:
            config.error("You must specify a distribution.")

        self.suppressions.append((None, re.compile(r"\.pbuilderrc does not exist"), None, None))

        self.addLogObserver('stdio', logobserver.LineConsumerLogObserver(self.logConsumer))

    @defer.inlineCallbacks
    def run(self):
        if self.basetgz is None:
            self.basetgz = self._default_basetgz
            kwargs = {}
            if self.architecture:
                kwargs['architecture'] = self.architecture
            else:
                kwargs['architecture'] = 'local'
            kwargs['distribution'] = self.distribution
            self.basetgz = self.basetgz.format(**kwargs)

        self.command = ['pdebuild', '--buildresult', '.', '--pbuilder', self.pbuilder]
        if self.architecture:
            self.command += ['--architecture', self.architecture]
        self.command += ['--', '--buildresult', '.', self.baseOption, self.basetgz]
        if self.extrapackages:
            self.command += ['--extrapackages', " ".join(self.extrapackages)]

        res = yield self.checkBasetgz()
        if res != results.SUCCESS:
            return res

        res = yield super().run()
        return res

    @defer.inlineCallbacks
    def checkBasetgz(self):
        cmd = remotecommand.RemoteCommand('stat', {'file': self.basetgz})
        yield self.runCommand(cmd)

        if cmd.rc != 0:
            log.msg("basetgz not found, initializing it.")

            command = [
                'sudo',
                self.pbuilder,
                '--create',
                self.baseOption,
                self.basetgz,
                '--distribution',
                self.distribution,
                '--mirror',
                self.mirror,
            ]
            if self.othermirror:
                command += ['--othermirror', self.othermirror]
            if self.architecture:
                command += ['--architecture', self.architecture]
            if self.extrapackages:
                command += ['--extrapackages', " ".join(self.extrapackages)]
            if self.keyring:
                command += ['--debootstrapopts', f"--keyring={self.keyring}"]
            if self.components:
                command += ['--components', self.components]

            cmd = remotecommand.RemoteShellCommand(self.workdir, command)

            stdio_log = yield self.addLog("pbuilder")
            cmd.useLog(stdio_log, True, "stdio")

            self.description = ["PBuilder", "create."]
            yield self.updateSummary()

            yield self.runCommand(cmd)
            if cmd.rc != 0:
                log.msg(f"Failure when running {cmd}.")
                return results.FAILURE
            return results.SUCCESS

        s = cmd.updates["stat"][-1]
        # basetgz will be a file when running in pbuilder
        # and a directory in case of cowbuilder
        if stat.S_ISREG(s[stat.ST_MODE]) or stat.S_ISDIR(s[stat.ST_MODE]):
            log.msg(f"{self.basetgz} found.")
            age = time.time() - s[stat.ST_MTIME]
            if age >= self.maxAge:
                log.msg("basetgz outdated, updating")
                command = ['sudo', self.pbuilder, '--update', self.baseOption, self.basetgz]

                cmd = remotecommand.RemoteShellCommand(self.workdir, command)
                stdio_log = yield self.addLog("pbuilder")
                cmd.useLog(stdio_log, True, "stdio")

                yield self.runCommand(cmd)
                if cmd.rc != 0:
                    log.msg(f"Failure when running {cmd}.")
                    return results.FAILURE
            return results.SUCCESS

        log.msg(f"{self.basetgz} is not a file or a directory.")
        return results.FAILURE

    def logConsumer(self):
        r = re.compile(r"dpkg-genchanges  >\.\./(.+\.changes)")
        while True:
            _, line = yield
            mo = r.search(line)
            if mo:
                self.setProperty("deb-changes", mo.group(1), "DebPbuilder")


class DebCowbuilder(DebPbuilder):
    """Build a debian package with cowbuilder inside of a chroot."""

    name = "cowbuilder"

    _default_basetgz = "/var/cache/pbuilder/{distribution}-{architecture}-buildbot.cow/"

    pbuilder = '/usr/sbin/cowbuilder'
    baseOption = '--basepath'


class UbuPbuilder(DebPbuilder):
    """Build a Ubuntu package with pbuilder inside of a chroot."""

    distribution = None
    mirror = "http://archive.ubuntu.com/ubuntu/"

    components = "main universe"


class UbuCowbuilder(DebCowbuilder):
    """Build a Ubuntu package with cowbuilder inside of a chroot."""

    distribution = None
    mirror = "http://archive.ubuntu.com/ubuntu/"

    components = "main universe"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from buildbot.data import base
from buildbot.data import types
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    from buildbot.db.codebases import CodebaseModel


def codebase_db_to_data(model: CodebaseModel) -> dict[str, Any]:
    return {
        "codebaseid": model.id,
        "name": model.name,
        "slug": model.slug,
        "projectid": model.projectid,
    }


codebases_field_map = {
    'codebaseid': 'codebases.id',
    'name': 'codebases.name',
    'slug': 'codebases.slug',
    'projectid': 'codebases.projectid',
}


class CodebaseEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/codebases/n:codebaseid",
        "/projects/n:projectid/codebases/i:codebasename",
    ]

    @async_to_deferred
    async def get(self, result_spec: base.ResultSpec, kwargs: Any) -> dict[str, Any] | None:
        result_spec.fieldMapping = codebases_field_map
        if 'codebaseid' in kwargs:
            codebase = await self.master.db.codebases.get_codebase(kwargs['codebaseid'])
            if not codebase:
                return None
            return codebase_db_to_data(codebase)

        codebases = await self.master.db.codebases.get_codebases(kwargs['projectid'])
        codebases = [c for c in codebases if c.name == kwargs['codebasename']]
        if not codebases:
            return None
        return codebase_db_to_data(codebases[0])


class CodebasesEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    rootLinkName = 'codebases'
    pathPatterns = [
        "/codebases",
    ]

    @async_to_deferred
    async def get(self, result_spec: base.ResultSpec, kwargs: Any) -> list[dict[str, Any]]:
        result_spec.fieldMapping = codebases_field_map
        codebases = await self.master.db.codebases.get_codebases(result_spec=result_spec)
        return [codebase_db_to_data(c) for c in codebases]


class Codebase(base.ResourceType):
    name = "codebase"
    plural = "codebases"
    endpoints = [CodebaseEndpoint, CodebasesEndpoint]
    eventPathPatterns = [
        "/codebases/:codebaseid",
    ]

    class EntityType(types.Entity):
        codebaseid = types.Integer()
        projectid = types.Integer()
        name = types.Identifier(70)
        slug = types.Identifier(70)

    entityType = EntityType(name)

    @async_to_deferred
    async def generate_event(self, _id: int, event: str) -> None:
        codebase = await self.master.data.get(('codebases', str(_id)))
        self.produceEvent(codebase, event)

    @base.updateMethod
    def find_codebase_id(
        self, *, projectid: int, name: str, auto_create: bool = True
    ) -> int | None:
        return self.master.db.codebases.find_codebase_id(
            projectid=projectid, name=name, auto_create=auto_create
        )

    @base.updateMethod
    @async_to_deferred
    async def update_codebase_info(
        self,
        *,
        codebaseid: int,
        projectid: int,
        slug: str,
    ) -> None:
        await self.master.db.codebases.update_codebase_info(
            codebaseid=codebaseid, projectid=projectid, slug=slug
        )
        await self.generate_event(codebaseid, "update")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os

import jinja2
from twisted.internet import defer
from twisted.python import util

from buildbot.config import master as config_master
from buildbot.master import BuildMaster
from buildbot.util import in_reactor


def makeBasedir(config):
    if os.path.exists(config['basedir']):
        if not config['quiet']:
            print("updating existing installation")
        return
    if not config['quiet']:
        print("mkdir", config['basedir'])
    os.mkdir(config['basedir'])


def makeTAC(config):
    # render buildbot_tac.tmpl using the config
    loader = jinja2.FileSystemLoader(os.path.dirname(__file__))
    env = jinja2.Environment(loader=loader, undefined=jinja2.StrictUndefined)
    env.filters['repr'] = repr
    tpl = env.get_template('buildbot_tac.tmpl')
    cxt = dict((k.replace('-', '_'), v) for k, v in config.items())
    contents = tpl.render(cxt)

    tacfile = os.path.join(config['basedir'], "buildbot.tac")
    if os.path.exists(tacfile):
        with open(tacfile, encoding='utf-8') as f:
            oldcontents = f.read()
        if oldcontents == contents:
            if not config['quiet']:
                print("buildbot.tac already exists and is correct")
            return
        if not config['quiet']:
            print("not touching existing buildbot.tac")
            print("creating buildbot.tac.new instead")
        tacfile += ".new"
    with open(tacfile, "w", encoding='utf-8') as f:
        f.write(contents)


def makeSampleConfig(config):
    source = util.sibpath(__file__, "sample.cfg")
    target = os.path.join(config['basedir'], "master.cfg.sample")
    if not config['quiet']:
        print(f"creating {target}")
    with open(source, encoding='utf-8') as f:
        config_sample = f.read()
    if config['db']:
        config_sample = config_sample.replace('sqlite:///state.sqlite', config['db'])
    with open(target, "w", encoding='utf-8') as f:
        f.write(config_sample)
    os.chmod(target, 0o600)


@defer.inlineCallbacks
def createDB(config):
    # create a master with the default configuration, but with db_url
    # overridden
    master_cfg = config_master.MasterConfig()
    master_cfg.db.db_url = config['db']
    master = BuildMaster(config['basedir'])
    master.config = master_cfg
    db = master.db
    yield db.setup(check_version=False, verbose=not config['quiet'])
    if not config['quiet']:
        print(f"creating database ({master_cfg.db.db_url})")
    yield db.model.upgrade()


@in_reactor
@defer.inlineCallbacks
def createMaster(config):
    makeBasedir(config)
    makeTAC(config)
    makeSampleConfig(config)
    yield createDB(config)

    if not config['quiet']:
        print(f"buildmaster configured in {config['basedir']}")

    return 0



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""add rebuilt_buildid column to buildsets table

Revision ID: 065
Revises: 064

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '065'
down_revision = '064'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("buildsets") as batch_op:
        batch_op.add_column(
            sa.Column(
                "rebuilt_buildid",
                sa.Integer,
                sa.ForeignKey(
                    "builds.id", use_alter=True, name="rebuilt_buildid", ondelete='SET NULL'
                ),
                nullable=True,
            ),
        )


def downgrade():
    op.drop_column("buildsets", "rebuilt_buildid")



#!/usr/bin/python

# This is a program which will poll a (remote) SVN repository, looking for
# new revisions. It then uses the 'buildbot sendchange' command to deliver
# information about the Change to a (remote) buildmaster. It can be run from
# a cron job on a periodic basis, or can be told (with the 'watch' option) to
# automatically repeat its check every 10 minutes.

# This script does not store any state information, so to avoid spurious
# changes you must use the 'watch' option and let it run forever.

# You will need to provide it with the location of the buildmaster's
# PBChangeSource port (in the form hostname:portnum), and the svnurl of the
# repository to watch.


import subprocess
import sys
import time
import xml.dom.minidom
from optparse import OptionParser
from xml.parsers.expat import ExpatError

if sys.platform == 'win32':
    import win32pipe


def getoutput(cmd):
    timeout = 120
    maxtries = 3
    if sys.platform == 'win32':
        f = win32pipe.popen(cmd)
        stdout = ''.join(f.readlines())
        f.close()
    else:
        currentry = 1
        while True:  # retry loop
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            waited = 0
            while True:  # wait loop
                if p.poll() != None:
                    break  # process ended.
                if waited > timeout:
                    print(
                        "WARNING: Timeout of {} seconds reached while trying to run: {}".format(
                            timeout, ' '.join(cmd)
                        )
                    )
                    break
                waited += 1
                time.sleep(1)

            if p.returncode != None:  # process has endend
                stdout = p.stdout.read()
                if p.returncode == 0:
                    break  # ok: exit retry loop
                else:
                    print(
                        'WARNING: "{}" returned status code: {}'.format(' '.join(cmd), p.returncode)
                    )
                    if stdout is not None:
                        print(stdout)
            else:
                p.kill()

            if currentry > maxtries:
                print(
                    "ERROR: Reached maximum number of tries ({}) to run: {}".format(
                        maxtries, ' '.join(cmd)
                    )
                )
                sys.exit(1)
            currentry += 1
    return stdout


def sendchange_cmd(master, revisionData):
    cmd = [
        "buildbot",
        "sendchange",
        f"--master={master}",
        "--revision={}".format(revisionData['revision']),
        "--who={}".format(revisionData['author']),
        "--comments={}".format(revisionData['comments']),
        "--vc={}".format('svn'),
    ]
    if opts.revlink:
        cmd.append("--revlink={}/{}".format(opts.revlink, revisionData['revision']))
    if opts.category:
        cmd.append(f"--category={opts.category}")
    if opts.branch:
        cmd.append(f"--branch={opts.branch}")
    if opts.auth:
        cmd.append(f"--auth={opts.auth}")
    for path in revisionData['paths']:
        cmd.append(path)

    if opts.verbose:
        print(cmd)

    return cmd


def parseChangeXML(raw_xml):
    """Parse the raw xml and return a dict with key pairs set.

    Commmand we're parsing:

    svn log --non-interactive --xml --verbose --limit=1 <repo url>

    With an output that looks like this:

    <?xml version="1.0"?>
     <log>
      <logentry revision="757">
       <author>mwiggins</author>
       <date>2009-11-11T17:16:48.012357Z</date>
       <paths>
        <path kind="" copyfrom-path="/trunk" copyfrom-rev="756" action="A">/tags/Latest</path>
       </paths>
       <msg>Updates/latest</msg>
      </logentry>
     </log>
    """

    data = dict()

    # parse the xml string and grab the first log entry.
    try:
        doc = xml.dom.minidom.parseString(raw_xml)
    except ExpatError:
        print("\nError: Got an empty response with an empty changeset.\n")
        raise
    log_entry = doc.getElementsByTagName("logentry")[0]

    # grab the appropriate meta data we need
    data['revision'] = log_entry.getAttribute("revision")
    data['author'] = "".join([
        t.data for t in log_entry.getElementsByTagName("author")[0].childNodes
    ])
    data['comments'] = "".join([
        t.data for t in log_entry.getElementsByTagName("msg")[0].childNodes
    ])

    # grab the appropriate file paths that changed.
    pathlist = log_entry.getElementsByTagName("paths")[0]
    paths = []
    if opts.branch:
        branchtoken = "/" + opts.branch.strip("/") + "/"
    for path in pathlist.getElementsByTagName("path"):
        filename = "".join([t.data for t in path.childNodes])
        if opts.branch:
            filename = filename.split(branchtoken, 1)[1]
        paths.append(filename)
    data['paths'] = paths

    return data


# FIXME: instead of just picking the last svn change each $interval minutes,
# we should be querying the svn server for all the changes between our
# last check and now, and notify the buildmaster about all of them.
# This is an example of a svn query we could do to get allo those changes:
# svn log --xml --non-interactive -r ${lastrevchecked}:HEAD https://repo.url/branch


def checkChanges(repo, master, oldRevision=-1):
    cmd = ["svn", "log", "--non-interactive", "--xml", "--verbose", "--limit=1", repo]

    if opts.verbose:
        print("Getting last revision of repository: " + repo)

    xml1 = getoutput(cmd)
    pretty_time = time.strftime("%F %T ")

    if opts.verbose:
        print("XML\n-----------\n" + xml1 + "\n\n")

    revisionData = parseChangeXML(xml1)

    if opts.verbose:
        print("PATHS")
        print(revisionData['paths'])

    if revisionData['revision'] != oldRevision:
        cmd = sendchange_cmd(master, revisionData)
        status = getoutput(cmd)

        print("{} Revision {}: {}".format(pretty_time, revisionData['revision'], status))

    else:
        print(
            "{} nothing has changed since revision {}".format(pretty_time, revisionData['revision'])
        )

    return revisionData['revision']


def build_parser():
    usagestr = "%prog [options] <repo url> <buildbot master:port>"
    parser = OptionParser(usage=usagestr)

    parser.add_option(
        "-c",
        "--category",
        dest="category",
        action="store",
        default="",
        help="""Store a category name to be associated with sendchange msg.""",
    )

    parser.add_option(
        "-i",
        "--interval",
        dest="interval",
        action="store",
        default=0,
        help="Implies watch option and changes the time in minutes to the value specified.",
    )

    parser.add_option(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="Enables more information to be presented on the command line.",
    )

    parser.add_option(
        "-b",
        "--branch",
        dest="branch",
        action="store",
        default=None,
        help="Watch only changes for this branch and send the branch info.",
    )

    parser.add_option(
        "-a",
        "--auth",
        dest="auth",
        action="store",
        default=None,
        help="Authentication token - username:password.",
    )

    parser.add_option(
        "-l",
        "--link",
        dest="revlink",
        action="store",
        default=None,
        help="A base URL for the revision links.",
    )

    parser.add_option(
        "",
        "--watch",
        dest="watch",
        action="store_true",
        default=False,
        help="Automatically check the repo url every 10 minutes.",
    )

    return parser


def validate_args(args):
    """Validate our arguments and exit if we don't have what we want."""
    if not args:
        print("\nError: No arguments were specified.\n")
        parser.print_help()
        sys.exit(1)
    elif len(args) > 2:
        print("\nToo many arguments specified.\n")
        parser.print_help()
        sys.exit(2)


if __name__ == '__main__':
    # build our parser and validate our args
    parser = build_parser()
    (opts, args) = parser.parse_args()
    validate_args(args)
    if opts.interval:
        try:
            int(opts.interval)
        except ValueError:
            print("\nError: Value of the interval option must be a number.")
            parser.print_help()
            sys.exit(3)

    # grab what we need
    repo_url = args[0]
    bbmaster = args[1]

    if opts.branch:
        repo_url = repo_url.rstrip("/") + "/" + opts.branch.lstrip("/")

    # if watch is specified, run until stopped
    if opts.watch or opts.interval:
        oldRevision = -1
        print(f"Watching for changes in repo {repo_url} for master {bbmaster}.")
        while True:
            try:
                oldRevision = checkChanges(repo_url, bbmaster, oldRevision)
            except ExpatError:
                # had an empty changeset.  Trapping the exception and moving
                # on.
                pass
            try:
                if opts.interval:
                    # Check the repository every interval in minutes the user
                    # specified.
                    time.sleep(int(opts.interval) * 60)
                else:
                    # Check the repository every 10 minutes
                    time.sleep(10 * 60)
            except KeyboardInterrupt:
                print("\nReceived interrupt via keyboard.  Shutting Down.")
                sys.exit(0)

    # default action if watch isn't specified
    checkChanges(repo_url, bbmaster)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from buildbot.db import base
from buildbot.util.sautils import hash_columns
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    from twisted.internet import defer


@dataclass
class ProjectModel:
    id: int
    name: str
    slug: str
    description: str | None
    description_format: str | None
    description_html: str | None

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'ProjectsConnectorComponent '
                'get_project, get_projects, and get_active_projects '
                'no longer return Project as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


class ProjectsConnectorComponent(base.DBConnectorComponent):
    def find_project_id(self, name: str, auto_create: bool = True) -> defer.Deferred[int | None]:
        name_hash = hash_columns(name)
        return self.findSomethingId(
            tbl=self.db.model.projects,
            whereclause=(self.db.model.projects.c.name_hash == name_hash),
            insert_values={
                "name": name,
                "slug": name,
                "name_hash": name_hash,
            },
            autoCreate=auto_create,
        )

    def get_project(self, projectid: int) -> defer.Deferred[ProjectModel | None]:
        def thd(conn) -> ProjectModel | None:
            q = self.db.model.projects.select().where(
                self.db.model.projects.c.id == projectid,
            )
            res = conn.execute(q)
            row = res.fetchone()

            rv = None
            if row:
                rv = self._model_from_row(row)
            res.close()
            return rv

        return self.db.pool.do(thd)

    def get_projects(self) -> defer.Deferred[list[ProjectModel]]:
        def thd(conn) -> list[ProjectModel]:
            tbl = self.db.model.projects
            q = tbl.select()
            q = q.order_by(tbl.c.name)
            res = conn.execute(q)
            return [self._model_from_row(row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    def get_active_projects(self) -> defer.Deferred[list[ProjectModel]]:
        def thd(conn) -> list[ProjectModel]:
            projects_tbl = self.db.model.projects
            builders_tbl = self.db.model.builders
            bm_tbl = self.db.model.builder_masters

            q = projects_tbl.select().join(builders_tbl).join(bm_tbl).order_by(projects_tbl.c.name)
            res = conn.execute(q)
            return [self._model_from_row(row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    # returns a Deferred that returns a value
    def update_project_info(
        self,
        projectid: int,
        slug: str,
        description: str | None,
        description_format: str | None,
        description_html: str | None,
    ) -> defer.Deferred[None]:
        def thd(conn) -> None:
            q = self.db.model.projects.update().where(self.db.model.projects.c.id == projectid)
            conn.execute(
                q.values(
                    slug=slug,
                    description=description,
                    description_format=description_format,
                    description_html=description_html,
                )
            ).close()

        return self.db.pool.do_with_transaction(thd)

    def _model_from_row(self, row):
        return ProjectModel(
            id=row.id,
            name=row.name,
            slug=row.slug,
            description=row.description,
            description_format=row.description_format,
            description_html=row.description_html,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import re

from twisted.trial import unittest

from buildbot.steps.shell import WarningCountingShellCommand


class TestWarningCountingShellCommand(unittest.TestCase):
    # Makes sure that it is possible to suppress warnings even if the
    # warning extractor does not provide line information

    def testSuppressingLinelessWarningsPossible(self):
        # Use a warningExtractor that does not provide line
        # information
        w = WarningCountingShellCommand(
            warningExtractor=WarningCountingShellCommand.warnExtractWholeLine, command="echo"
        )

        # Add suppression manually instead of using suppressionFile
        fileRe = None
        warnRe = ".*SUPPRESS.*"
        start = None
        end = None
        suppression = (fileRe, warnRe, start, end)
        w.addSuppression([suppression])

        # Now call maybeAddWarning
        warnings = []
        line = "this warning should be SUPPRESSed"
        match = re.match(".*warning.*", line)
        w.maybeAddWarning(warnings, line, match)

        # Finally make the suppressed warning was *not* added to the
        # list of warnings
        expectedWarnings = 0
        self.assertEqual(len(warnings), expectedWarnings)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members
# Portions Copyright 2010 Isotoma Limited

import os
import socket

from twisted.internet import defer
from twisted.python import log

from buildbot import config
from buildbot.interfaces import LatentWorkerFailedToSubstantiate
from buildbot.util import runprocess
from buildbot.util.queue import ConnectableThreadQueue
from buildbot.worker import AbstractLatentWorker

try:
    import libvirt
except ImportError:
    libvirt = None


def handle_connect_close(conn, reason, opaque):
    opaque.close_connection()


class ThreadWithQueue(ConnectableThreadQueue):
    def __init__(self, pool, uri, *args, **kwargs):
        self.pool = pool  # currently used only for testing
        self.uri = uri
        super().__init__(*args, **kwargs)

    def on_close_connection(self, conn):
        self.close_connection()

    def close_connection(self):
        conn = self.conn
        super().close_connection()
        conn.close()

    def libvirt_open(self):
        return libvirt.open(self.uri)

    def create_connection(self):
        try:
            log.msg(f"Connecting to {self.uri}")
            conn = self.libvirt_open()
            conn.registerCloseCallback(handle_connect_close, self)
            log.msg(f"Connected to {self.uri}")
            return conn
        except Exception as e:
            log.err(f"Error connecting to {self.uri}: {e}, will retry later")
            return None


class ServerThreadPool:
    ThreadClass = ThreadWithQueue

    def __init__(self):
        self.threads = {}

    def do(self, uri, func, *args, **kwargs):
        # returns a Deferred
        if uri not in self.threads:
            self.threads[uri] = self.ThreadClass(self, uri)

        def logging_func(conn, *args, **kwargs):
            try:
                return func(conn, *args, **kwargs)
            except Exception as e:
                log.err(f"libvirt: Exception on {uri}: {e}")
                raise

        return self.threads[uri].execute_in_thread(logging_func, *args, **kwargs)

    def is_connected(self, uri):
        if uri in self.threads:
            return self.threads[uri].conn is not None
        return False

    def is_connecting(self, uri):
        if uri in self.threads:
            return self.threads[uri].connecting
        return False

    @defer.inlineCallbacks
    def get_or_create_connection(self, uri):
        if uri not in self.threads:
            yield self.do(uri, lambda: None)
        return self.threads[uri].conn

    def reset_connection(self, uri):
        if uri in self.threads:
            self.threads[uri].close_connection()
        else:
            log.err(f'libvirt.ServerThreadPool: Unknown connection {uri}')


# A module is effectively a singleton class, so this is OK
threadpool = ServerThreadPool()


class Connection:
    def __init__(self, uri):
        self.uri = uri


class LibVirtWorker(AbstractLatentWorker):
    pool = threadpool
    metadata = '<auth username="{}" password="{}" master="{}"/>'
    ns = 'http://buildbot.net/'
    metakey = 'buildbot'

    def __init__(
        self,
        name,
        password,
        hd_image=None,
        base_image=None,
        uri="system:///",
        xml=None,
        masterFQDN=None,
        **kwargs,
    ):
        super().__init__(name, password, **kwargs)
        if not libvirt:
            config.error("The python module 'libvirt' is needed to use a LibVirtWorker")

        self.uri = uri
        self.image = hd_image
        self.base_image = base_image
        self.xml = xml
        if masterFQDN:
            self.masterFQDN = masterFQDN
        else:
            self.masterFQDN = socket.getfqdn()

        self.cheap_copy = True
        self.graceful_shutdown = False

    def _pool_do(self, func):
        return self.pool.do(self.uri, func)

    @defer.inlineCallbacks
    def _get_domain(self):
        try:
            domain = yield self._pool_do(lambda conn: conn.lookupByName(self.workername))
            return domain
        except libvirt.libvirtError as e:
            log.err(f'LibVirtWorker: got error when accessing domain: {e}')
            try:
                self.pool.reset_connection(self.uri)
            except Exception as e1:
                log.err(f'LibVirtWorker: got error when resetting connection: {e1}')
            raise e

    @defer.inlineCallbacks
    def _get_domain_id(self):
        domain = yield self._get_domain()
        if domain is None:
            return -1
        domain_id = yield self._pool_do(lambda conn: domain.ID())
        return domain_id

    @defer.inlineCallbacks
    def _prepare_base_image(self):
        """
        I am a private method for creating (possibly cheap) copies of a
        base_image for start_instance to boot.
        """
        if not self.base_image:
            return

        if self.cheap_copy:
            clone_cmd = [
                'qemu-img',
                'create',
                '-o',
                'backing_fmt=qcow2',
                '-b',
                self.base_image,
                '-f',
                'qcow2',
                self.image,
            ]
        else:
            clone_cmd = ['cp', self.base_image, self.image]

        log.msg(f"Cloning base image: {clone_cmd}'")

        try:
            rc = yield runprocess.run_process(
                self.master.reactor, clone_cmd, collect_stdout=False, collect_stderr=False
            )
            if rc != 0:
                raise LatentWorkerFailedToSubstantiate(f'Failed to clone image (rc={rc})')
        except Exception as e:
            log.err(f"Cloning failed: {e}")
            raise

    @defer.inlineCallbacks
    def start_instance(self, build):
        """
        I start a new instance of a VM.

        If a base_image is specified, I will make a clone of that otherwise i will
        use image directly.

        If i'm not given libvirt domain definition XML, I will look for my name
        in the list of defined virtual machines and start that.
        """

        try:
            domain_id = yield self._get_domain_id()
            if domain_id != -1:
                raise LatentWorkerFailedToSubstantiate(
                    f"{self}: Cannot start_instance as it's already active"
                )
        except Exception as e:
            raise LatentWorkerFailedToSubstantiate(
                f'{self}: Got error while retrieving domain ID: {e}'
            ) from e

        yield self._prepare_base_image()

        try:
            if self.xml:
                yield self._pool_do(lambda conn: conn.createXML(self.xml, 0))
            else:
                domain = yield self._get_domain()
                yield self._pool_do(
                    lambda conn: domain.setMetadata(
                        libvirt.VIR_DOMAIN_METADATA_ELEMENT,
                        self.metadata.format(self.workername, self.password, self.masterFQDN),
                        self.metakey,
                        self.ns,
                        libvirt.VIR_DOMAIN_AFFECT_CONFIG,
                    )
                )

                yield self._pool_do(lambda conn: domain.create())

        except Exception as e:
            raise LatentWorkerFailedToSubstantiate(
                f'{self}: Got error while starting VM: {e}'
            ) from e

        return True

    @defer.inlineCallbacks
    def stop_instance(self, fast=False):
        """
        I attempt to stop a running VM.
        I make sure any connection to the worker is removed.
        If the VM was using a cloned image, I remove the clone
        When everything is tidied up, I ask that bbot looks for work to do
        """

        domain_id = yield self._get_domain_id()
        if domain_id == -1:
            log.msg(f"{self}: Domain is unexpectedly not running")
            return

        domain = yield self._get_domain()

        if self.graceful_shutdown and not fast:
            log.msg(f"Graceful shutdown chosen for {self.workername}")
            try:
                yield self._pool_do(lambda conn: domain.shutdown())
            except Exception as e:
                log.msg(f'{self}: Graceful shutdown failed ({e}). Force destroying domain')
                # Don't re-throw to stop propagating shutdown error if destroy was successful.
                yield self._pool_do(lambda conn: domain.destroy())

        else:
            yield self._pool_do(lambda conn: domain.destroy())

        if self.base_image:
            log.msg(f'{self}: Removing image {self.image}')
            os.remove(self.image)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import copy
import datetime
import hashlib
import json
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Pattern
from typing import Sequence

from twisted.internet import defer
from twisted.python import log

from buildbot import config
from buildbot import util
from buildbot.changes import base
from buildbot.changes.filter import ChangeFilter
from buildbot.config.checks import check_param_int
from buildbot.config.checks import check_param_int_none
from buildbot.util import bytes2unicode
from buildbot.util import deferwaiter
from buildbot.util import httpclientservice
from buildbot.util import runprocess
from buildbot.util import watchdog
from buildbot.util.protocol import LineProcessProtocol
from buildbot.util.pullrequest import PullRequestMixin

if TYPE_CHECKING:
    from buildbot.util.twisted import InlineCallbacksType


def _canonicalize_event(event: dict) -> dict:
    """
    Return an event dictionary which is consistent between the gerrit
    event stream and the gerrit event log formats.
    """
    # For "patchset-created" the events-log JSON looks like:
    #   "project": {"name": "buildbot"}
    # while the stream-events JSON looks like:
    #   "project": "buildbot"
    # so we canonicalize them to the latter
    if "change" not in event:
        return event

    change = event["change"]
    if "project" not in change:
        return event

    project = change["project"]
    if not isinstance(project, dict):
        return event

    if "name" not in project:
        return event

    event = copy.deepcopy(event)
    event["change"]["project"] = project["name"]
    return event


class GerritChangeFilter(ChangeFilter):
    """This gerrit specific change filter helps creating pre-commit and post-commit builders"""

    compare_attrs: ClassVar[Sequence[str]] = ('eventtype_fn', 'gerrit_branch_fn')

    def __init__(
        self,
        branch: Any = util.NotABranch,
        branch_re: str | Pattern | None = None,
        branch_fn: Callable[[str], bool] | None = None,
        eventtype: str | None = None,
        eventtype_re: str | Pattern | None = None,
        eventtype_fn: Callable[[str], bool] | None = None,
        **kw: Any,
    ):
        if eventtype is not None:
            kw.setdefault('property_eq', {})['event.type'] = eventtype
        if eventtype_re is not None:
            kw.setdefault('property_re', {})['event.type'] = eventtype_re

        # for branch change filter, we take the real gerrit branch
        # instead of the change's branch, which is also used as a grouping key
        if branch is not util.NotABranch:
            kw.setdefault('property_eq', {})['event.change.branch'] = branch
        if branch_re is not None:
            kw.setdefault('property_re', {})['event.change.branch'] = branch_re

        super().__init__(**kw)
        self.eventtype_fn = eventtype_fn
        self.gerrit_branch_fn = branch_fn

    def filter_change(self, change: Any) -> bool:
        if self.eventtype_fn is not None:
            value = change.properties.getProperty('event.type', '')
            if not self.eventtype_fn(value):
                return False

        if self.gerrit_branch_fn is not None:
            # for branch change filter, we take the real gerrit branch
            # instead of the change's branch, which is also used as a grouping key
            value = change.properties.getProperty('event.change.branch', '')
            if not self.gerrit_branch_fn(value):
                return False

        return super().filter_change(change)

    def _get_repr_filters(self) -> list[str]:
        filters = super()._get_repr_filters()
        if self.eventtype_fn is not None:
            filters.append(f'{self.eventtype_fn.__name__}(eventtype)')
        if self.gerrit_branch_fn is not None:
            filters.append(f'{self.gerrit_branch_fn.__name__}(branch)')
        return filters


def _gerrit_user_to_author(props: dict, username: str = "unknown") -> str:
    """
    Convert Gerrit account properties to Buildbot format

    Take into account missing values
    """
    username = props.get("username", username)
    username = props.get("name", username)
    if "email" in props:
        username += f" <{props['email']}>"
    return username


class GerritChangeSourceBase(base.ChangeSource, PullRequestMixin):
    """This source will maintain a connection to gerrit ssh server
    that will provide us gerrit events in json format."""

    compare_attrs: ClassVar[Sequence[str]] = ("gerritserver", "gerritport")
    name: str | None = None  # type: ignore[assignment]
    # list of properties that are no of no use to be put in the event dict
    external_property_denylist = ["event.eventCreatedOn"]
    external_property_whitelist = ['*']
    property_basename = 'event'

    def checkConfig(
        self,
        gitBaseURL: str | None = None,
        handled_events: tuple[str, ...] = ("patchset-created", "ref-updated"),
        debug: bool = False,
        get_files: bool = False,
    ) -> None:  # type: ignore[override]  # checkConfig doesn't need to match supertype
        if gitBaseURL is None:
            config.error("gitBaseURL must be specified")

    def reconfigService(
        self,
        gitBaseURL: str | None = None,
        handled_events: tuple[str, ...] = ("patchset-created", "ref-updated"),
        debug: bool = False,
        get_files: bool = False,
    ) -> None:  # type: ignore[override]  # reconfigService doesn't need to match supertype
        self.gitBaseURL = gitBaseURL
        self.handled_events = list(handled_events)
        self._get_files = get_files
        self.debug = debug

    def build_properties(self, event: dict) -> dict:
        properties = self.extractProperties(event)
        properties["event.source"] = self.__class__.__name__
        if event['type'] in ('patchset-created', 'comment-added') and 'change' in event:
            properties['target_branch'] = event["change"]["branch"]
        return properties

    def getFiles(self, change: str, patchset: str) -> defer.Deferred:
        raise NotImplementedError

    def eventReceived(self, event: dict) -> defer.Deferred:
        if event['type'] not in self.handled_events:
            if self.debug:
                log.msg(f"the event type '{event['type']}' is not setup to handle")
            return defer.succeed(None)

        properties = self.build_properties(event)
        func_name = f'eventReceived_{event["type"].replace("-", "_")}'
        func = getattr(self, func_name, None)
        if func is None:
            return self.addChangeFromEvent(properties, event)

        return func(properties, event)

    def get_branch_from_event(self, event: dict) -> str:
        if event['type'] in ('patchset-created', 'comment-added'):
            return event["patchSet"]["ref"]
        return event["change"]["branch"]

    def strip_refs_heads_from_branch(self, branch: str) -> str:
        if branch.startswith('refs/heads/'):
            branch = branch[len('refs/heads/') :]
        return branch

    @defer.inlineCallbacks
    def addChangeFromEvent(self, properties: dict, event: dict) -> InlineCallbacksType[None]:
        if "change" not in event:
            if self.debug:
                log.msg(f'unsupported event {event["type"]}')
            return None

        if "patchSet" not in event:
            if self.debug:
                log.msg(f'unsupported event {event["type"]}')
            return None

        event = _canonicalize_event(event)
        event_change = event["change"]

        files = ["unknown"]
        if self._get_files:
            files = yield self.getFiles(
                change=event_change["number"], patchset=event["patchSet"]["number"]
            )

        yield self.master.data.updates.addChange(
            author=_gerrit_user_to_author(event_change["owner"]),
            project=util.bytes2unicode(event_change["project"]),
            repository=f'{self.gitBaseURL}/{event_change["project"]}',
            branch=self.get_branch_from_event(event),
            revision=event["patchSet"]["revision"],
            revlink=event_change["url"],
            comments=event_change["subject"],
            files=files,
            category=event["type"],
            properties=properties,
        )
        return

    def eventReceived_ref_updated(self, properties: dict, event: dict) -> defer.Deferred:
        ref = event["refUpdate"]
        author = "gerrit"

        if "submitter" in event:
            author = _gerrit_user_to_author(event["submitter"], author)

        # Ignore ref-updated events if patchset-created events are expected for this push.
        # ref-updated events may arrive before patchset-created events and cause problems, as
        # builds would be using properties from ref-updated event and not from patchset-created.
        # As a result it may appear that the change was not related to a Gerrit change and cause
        # reporters to not submit reviews for example.
        if 'patchset-created' in self.handled_events and ref['refName'].startswith('refs/changes/'):
            return defer.succeed(None)

        return self.master.data.updates.addChange(
            author=author,
            project=ref["project"],
            repository=f'{self.gitBaseURL}/{ref["project"]}',
            branch=self.strip_refs_heads_from_branch(ref["refName"]),
            revision=ref["newRev"],
            comments="Gerrit: commit(s) pushed.",
            files=["unknown"],
            category=event["type"],
            properties=properties,
        )


class GerritSshStreamEventsConnector:
    class LocalPP(LineProcessProtocol):
        MAX_STORED_OUTPUT_DEBUG_LINES = 20

        def __init__(self, connector: GerritSshStreamEventsConnector):
            super().__init__()
            self.connector = connector
            self._output_enabled = True
            self._ended_deferred: defer.Deferred[None] = defer.Deferred()

        @defer.inlineCallbacks
        def outLineReceived(self, line: bytes) -> InlineCallbacksType[None]:
            if self.connector.debug:
                log.msg(
                    f"{self.connector.change_source.name} "
                    + f"stdout: {line.decode('utf-8', errors='replace')}"
                )

            self.connector._append_line_for_debug(line)
            if self._output_enabled:
                yield self.connector.on_line_received_cb(line)

        def errLineReceived(self, line: bytes) -> None:
            if self.connector.debug:
                log.msg(
                    f"{self.connector.change_source.name} "
                    + f"stderr: {line.decode('utf-8', errors='replace')}"
                )
            if self._output_enabled:
                self.connector._append_line_for_debug(line)

        def processEnded(self, status: Any) -> None:
            super().processEnded(status)
            self._ended_deferred.callback(None)
            self.connector._stream_process_stopped()

        def disable_output(self) -> None:
            self._output_enabled = False

        def wait(self) -> defer.Deferred:
            return self._ended_deferred

    # (seconds) connections longer than this are considered good, and reset the backoff timer
    STREAM_GOOD_CONNECTION_TIME = 120

    # (seconds) minimum, but nonzero, time to wait before retrying a failed connection
    STREAM_BACKOFF_MIN = 0.5

    # multiplier used to increase the backoff from MIN to MAX on repeated failures
    STREAM_BACKOFF_EXPONENT = 1.5

    # (seconds) maximum time to wait before retrying a failed connection
    STREAM_BACKOFF_MAX = 60

    # The number of gerrit output lines to print in case of a failure
    MAX_STORED_OUTPUT_DEBUG_LINES = 20

    debug = False

    def __init__(
        self,
        reactor: Any,
        change_source: Any,
        gerritserver: str,
        username: str,
        *,
        gerritport: int = 29418,
        identity_file: str | None = None,
        ssh_server_alive_interval_s: int | None = 15,
        ssh_server_alive_count_max: int | None = 3,
        on_process_start_cb: Callable[[], None],
        on_line_received_cb: Callable[[bytes], defer.Deferred],
    ):
        self.reactor = reactor
        self.change_source = change_source
        self.gerritserver = gerritserver
        self.username = username
        self.gerritport = gerritport
        self.identity_file = identity_file
        self.ssh_server_alive_interval_s = ssh_server_alive_interval_s
        self.ssh_server_alive_count_max = ssh_server_alive_count_max
        self.on_process_start_cb = on_process_start_cb
        self.on_line_received_cb = on_line_received_cb
        self._process: tuple[Any, Any] | None = None
        self._stream_process_timeout = self.STREAM_BACKOFF_MIN
        self._last_lines_for_debug: list[bytes] = []

    def start(self) -> None:
        self._want_process = True
        self.start_stream_process()

    @defer.inlineCallbacks
    def stop(self) -> InlineCallbacksType[None]:
        self._want_process = False
        if self._process is not None:
            self._process[0].disable_output()
            self._process[1].signalProcess("KILL")
            yield self._process[0].wait()

    @defer.inlineCallbacks
    def restart(self) -> InlineCallbacksType[None]:
        if self._process is not None:
            self._process[0].disable_output()
            # Process will restart automatically
            self._process[1].signalProcess("KILL")
            yield self._process[0].wait()
        else:
            self.start()

    def _append_line_for_debug(self, line: bytes) -> None:
        self._last_lines_for_debug.append(line)
        while len(self._last_lines_for_debug) > self.MAX_STORED_OUTPUT_DEBUG_LINES:
            self._last_lines_for_debug.pop(0)

    def _build_gerrit_command(self, *gerrit_args: str) -> list[str]:
        """Get an ssh command list which invokes gerrit with the given args on the
        remote host"""

        options = [
            "-o",
            "BatchMode=yes",
        ]
        if self.ssh_server_alive_interval_s is not None:
            options += ["-o", f"ServerAliveInterval={self.ssh_server_alive_interval_s}"]
        if self.ssh_server_alive_count_max is not None:
            options += ["-o", f"ServerAliveCountMax={self.ssh_server_alive_count_max}"]

        cmd = ["ssh", *options, f"{self.username}@{self.gerritserver}", "-p", str(self.gerritport)]

        if self.identity_file is not None:
            cmd.extend(["-i", self.identity_file])

        cmd.append("gerrit")
        cmd.extend(gerrit_args)
        return cmd

    def start_stream_process(self) -> None:
        if self._process is not None:
            return

        if self.debug:
            log.msg(f"{self.change_source.name}: starting 'gerrit stream-events'")

        # Must be called before start of the process to ensure consistent ordering to avoid race
        # conditions.
        self.on_process_start_cb()

        cmd = self._build_gerrit_command("stream-events")
        self._last_stream_process_start = self.reactor.seconds()
        protocol = self.LocalPP(self)
        self._process = (protocol, self.reactor.spawnProcess(protocol, "ssh", cmd, env=None))
        self._last_lines_for_debug = []

    def _stream_process_stopped(self) -> None:
        self._process = None

        # if the service is stopped, don't try to restart the process
        if not self._want_process or not self.change_source.running:
            return

        now = self.reactor.seconds()
        if now - self._last_stream_process_start < self.STREAM_GOOD_CONNECTION_TIME:
            # bad startup; start the stream process again after a timeout,
            # and then increase the timeout
            log_lines = "\n".join([
                l.decode("utf-8", errors="ignore") for l in self._last_lines_for_debug
            ])

            log.msg(
                f"{self.change_source.name}: stream-events failed; restarting after "
                f"{round(self._stream_process_timeout)}s.\n"
                f"{len(self._last_lines_for_debug)} log lines follow:\n{log_lines}"
            )

            self.reactor.callLater(self._stream_process_timeout, self.start_stream_process)
            self._stream_process_timeout *= self.STREAM_BACKOFF_EXPONENT
            self._stream_process_timeout = min(
                self._stream_process_timeout, self.STREAM_BACKOFF_MAX
            )
        else:
            # good startup, but lost connection; restart immediately,
            # and set the timeout to its minimum

            # make sure we log the reconnection, so that it might be detected
            # and network connectivity fixed
            log.msg(f"{self.change_source.name}: stream-events lost connection. Reconnecting...")
            self.start_stream_process()
            self._stream_process_timeout = self.STREAM_BACKOFF_MIN

    @defer.inlineCallbacks
    def get_files(self, change: str, patchset: str) -> InlineCallbacksType[list[str]]:
        cmd = self._build_gerrit_command(
            "query", str(change), "--format", "JSON", "--files", "--patch-sets"
        )

        if self.debug:
            log.msg(
                f"{self.change_source.name}: querying for changed files in change {change}/{patchset}: {cmd}"
            )

        rc, out = yield runprocess.run_process(self.reactor, cmd, env=None, collect_stderr=False)
        if rc != 0:
            return ["unknown"]

        out = out.splitlines()[0]
        res = json.loads(bytes2unicode(out))

        if res.get("rowCount") == 0:
            return ["unknown"]

        patchsets = {i["number"]: i["files"] for i in res["patchSets"]}
        return [i["file"] for i in patchsets[int(patchset)]]


class GerritHttpEventLogPollerConnector:
    FIRST_FETCH_LOOKBACK_DAYS = 30

    debug = False

    def __init__(
        self,
        reactor: Any,
        change_source: Any,
        base_url: str,
        auth: Any,
        get_last_event_ts: Callable[[], defer.Deferred[int] | int],
        first_fetch_lookback: int,
        on_lines_received_cb: Callable[[list[bytes]], defer.Deferred],
    ) -> None:
        if base_url.endswith('/'):
            base_url = base_url[:-1]
        self._reactor = reactor
        self._change_source = change_source
        self._get_last_event_ts = get_last_event_ts
        self._base_url = base_url
        self._auth = auth
        self._first_fetch_lookback = first_fetch_lookback
        self._on_lines_received_cb = on_lines_received_cb
        self._last_event_time = None

    @defer.inlineCallbacks
    def setup(self) -> InlineCallbacksType[None]:
        self._http = yield httpclientservice.HTTPSession(
            self._change_source.master.httpservice, self._base_url, auth=self._auth
        )

    @defer.inlineCallbacks
    def poll(self) -> InlineCallbacksType[None]:
        last_event_ts = yield self._get_last_event_ts()
        if last_event_ts is None:
            # If there is not last event time stored in the database, then set
            # the last event time to some historical look-back
            last_event = datetime.datetime.fromtimestamp(
                self._reactor.seconds(), datetime.timezone.utc
            ) - datetime.timedelta(days=self._first_fetch_lookback)
        else:
            last_event = datetime.datetime.fromtimestamp(last_event_ts, datetime.timezone.utc)
        last_event_formatted = last_event.strftime("%Y-%m-%d %H:%M:%S")

        if self.debug:
            log.msg(f"{self._change_source.name}: Polling gerrit: {last_event_formatted}")

        res = yield self._http.get(
            "/plugins/events-log/events/", params={"t1": last_event_formatted}
        )
        if res.code != 200:
            log.msg(f'{self._change_source.name}: Polling gerrit: got HTTP error code {res.code}')
            return

        lines = yield res.content()
        yield self._on_lines_received_cb(lines.splitlines())

    @defer.inlineCallbacks
    def get_files(self, change: str, patchset: str) -> InlineCallbacksType[list[str]]:
        res = yield self._http.get(f"/changes/{change}/revisions/{patchset}/files/")
        res = yield res.content()
        try:
            res = res.splitlines()[1].decode('utf8')  # the first line of every response is `)]}'`
            return list(json.loads(res))
        except Exception as e:
            log.err(e, 'while getting files from connector')
            return []

    @defer.inlineCallbacks
    def do_poll(self) -> InlineCallbacksType[None]:
        try:
            yield self.poll()
        except Exception as e:
            log.err(e, 'while polling for changes')


def extract_gerrit_event_time(event: dict) -> int:
    return event["eventCreatedOn"]


def build_gerrit_event_hash(event: dict) -> str:
    return hashlib.sha1(json.dumps(event, sort_keys=True).encode("utf-8")).hexdigest()


def is_event_valid(event: Any) -> bool:
    return isinstance(event, dict) and "type" in event and "eventCreatedOn" in event


class GerritChangeSource(GerritChangeSourceBase):
    """This source will maintain a connection to gerrit ssh server that will provide us gerrit
    events in json format. Additionally, connection to gerrit HTTP server may be setup so
    that historical events can be fetched to fill any gaps due to Buildbot or Gerrit restarts
    or internet connectivity problems.

    Important considerations for filling gaps in processed events:
     - Gerrit events do not have unique IDs, only eventCreateOn timestamp which is common between
       events coming from the HTTP and SSH APIs
     - Gerrit HTTP API does not provide any ordering guarantees.
     - Gerrit HTTP and SSH APIs return events encoded identically
    """

    compare_attrs: ClassVar[Sequence[str]] = ("gerritserver", "gerritport")

    name = None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._stream_connector: GerritSshStreamEventsConnector | None = None
        self._poll_connector: GerritHttpEventLogPollerConnector | None = None

        self._queued_stream_events: list[tuple[int, dict]] = []

        # Events are received from stream event source continuously. If HTTP API is not available,
        # GerritChangeSource is always in this state.
        #
        # This variable is used to synchronize between concurrent data ingestion from poll or
        # stream event sources. If _is_synchronized == True, then polling data is discarded.
        # Otherwise, data from stream data source goes into _queued_stream_events.
        self._is_synchronized = True

        # True if SSH stream did not get events for a long time. It is unclear whether the
        # connection broke or there were simply no activity, so the SSH connection should not be
        # restarted. Note that _stream_activity_watchdog is disabled when
        # _stream_messages_timeout is True and should be re-enabled when exiting out of this
        # condition.
        self._stream_messages_timeout = False

        # Used for polling if last event timestamp is unknown.
        self._start_ts: int | None = None

        # Stores newest events that have been published for further processing and have identical
        # timestamp. This is used to ensure that events are not duplicated across stream and
        # polled sources.
        self._last_second_events: list[dict] = []
        # Contains hashes of self._last_second_events coming from previous run of this service.
        # self._last_second_events is not stored directly because of size considerations.
        self._last_second_event_hashes: list[str] | None = []

        self._last_event_ts: int | None = None
        # Last event timestamp recorded to database. Equivalent to self._last_event_ts. Separate
        # variable is support single database transaction for message batches.
        self._last_event_ts_saved: int | None = None

        self._deferwaiter = deferwaiter.DeferWaiter()
        self._poll_handler: deferwaiter.NonRepeatedActionHandler | None = None
        self._stream_activity_watchdog: watchdog.Watchdog | None = None

    def checkConfig(  # type: ignore[override]
        self,
        gerritserver: str,
        username: str,
        gerritport: int = 29418,
        identity_file: str | None = None,
        ssh_server_alive_interval_s: int | None = 15,
        ssh_server_alive_count_max: int | None = 3,
        http_url: str | None = None,
        http_auth: Any = None,
        http_poll_interval: int = 30,
        **kwargs: Any,
    ) -> None:
        if self.name is None:
            self.name = f"GerritChangeSource:{username}@{gerritserver}:{gerritport}"
        if 'gitBaseURL' not in kwargs:
            kwargs['gitBaseURL'] = "automatic at reconfigure"
        check_param_int_none(
            ssh_server_alive_interval_s, self.__class__, "ssh_server_alive_interval_s"
        )
        check_param_int_none(
            ssh_server_alive_count_max, self.__class__, "ssh_server_alive_count_max"
        )
        check_param_int(http_poll_interval, self.__class__, "http_poll_interval")
        super().checkConfig(**kwargs)

    @defer.inlineCallbacks
    def reconfigService(  # type: ignore[override]
        self,
        gerritserver: str,
        username: str,
        gerritport: int = 29418,
        identity_file: str | None = None,
        name: str | None = None,
        ssh_server_alive_interval_s: int | None = 15,
        ssh_server_alive_count_max: int | None = 3,
        http_url: str | None = None,
        http_auth: Any = None,
        http_poll_interval: int = 30,
        **kwargs: Any,
    ) -> InlineCallbacksType[None]:
        if 'gitBaseURL' not in kwargs:
            kwargs['gitBaseURL'] = f"ssh://{username}@{gerritserver}:{gerritport}"
        self.gerritserver = gerritserver
        self.gerritport = gerritport
        self.username = username
        self.identity_file = identity_file
        self._http_poll_interval = http_poll_interval

        if self._stream_connector is None:
            # TODO: this does not support reconfiguration at the moment
            self._stream_connector = GerritSshStreamEventsConnector(
                self.master.reactor,
                self,
                gerritserver,
                username,
                gerritport=gerritport,
                identity_file=identity_file,
                ssh_server_alive_interval_s=ssh_server_alive_interval_s,
                ssh_server_alive_count_max=ssh_server_alive_count_max,
                on_process_start_cb=self._stream_process_started,
                on_line_received_cb=self._line_received_stream,
            )
        self._stream_messages_timeout = False

        self._oid = yield self.master.db.state.getObjectId(self.name, self.__class__.__name__)

        if self._start_ts is None:
            self._start_ts = self.master.reactor.seconds()

        if self._last_event_ts is None:
            self._last_event_ts = yield self.master.db.state.getState(
                self._oid, 'last_event_ts', None
            )
            self._last_second_event_hashes = yield self.master.db.state.getState(
                self._oid, "last_event_hashes", None
            )

        if self._poll_handler is not None:
            self._poll_handler.stop()
        self._poll_handler = deferwaiter.NonRepeatedActionHandler(
            self.master.reactor,
            self._deferwaiter,
            lambda: self._poll_connector.do_poll(),  # type: ignore[union-attr]
        )

        if http_url is not None:
            if self._poll_connector is None:
                # TODO: this does not support reconfiguration at the moment
                self._poll_connector = GerritHttpEventLogPollerConnector(
                    self.master.reactor,
                    self,
                    http_url + "/a",
                    http_auth,
                    lambda: self._last_event_ts or self._start_ts or 0,  # 0 can't happen
                    first_fetch_lookback=0,
                    on_lines_received_cb=self._lines_received_poll,
                )
                yield self._poll_connector.setup()  # type: ignore[attr-defined]
            self._is_synchronized = False
        else:
            self._poll_connector = None
            self._is_synchronized = True

        self._stream_activity_watchdog = watchdog.Watchdog(
            self.master.reactor, self._no_stream_activity_timed_out, self._http_poll_interval
        )

        yield super().reconfigService(**kwargs)

    def activate(self) -> defer.Deferred[None]:
        assert self._stream_connector is not None
        assert self._stream_activity_watchdog is not None

        self._is_synchronized = self._poll_connector is None
        self._stream_connector.start()
        self._stream_activity_watchdog.start()
        return defer.succeed(None)

    @defer.inlineCallbacks
    def deactivate(self) -> InlineCallbacksType[None]:
        assert self._stream_connector is not None
        assert self._poll_handler is not None
        assert self._stream_activity_watchdog is not None

        self._stream_activity_watchdog.stop()
        yield self._stream_connector.stop()  # Note that this immediately stops event acceptance
        self._poll_handler.stop()
        yield self._deferwaiter.wait()

        if self._last_second_events:
            yield self.master.db.state.setState(
                self._oid,
                "last_event_hashes",
                [build_gerrit_event_hash(event) for event in self._last_second_events],
            )
        else:
            yield self.master.db.state.setState(self._oid, "last_event_hashes", None)

    def getFiles(self, change: str, patchset: str) -> defer.Deferred:
        assert self._stream_connector is not None
        return self._stream_connector.get_files(change, patchset)

    def _no_stream_activity_timed_out(self) -> None:
        assert self._poll_handler is not None
        if self._poll_connector is None:
            return
        self._stream_messages_timeout = True
        self._poll_handler.force()

    def _stream_process_started(self) -> None:
        assert self._poll_handler is not None
        if self._poll_connector is None:
            return
        self._is_synchronized = False
        self._poll_handler.force()

    def _record_last_second_event(self, event: dict, ts: int) -> None:
        if self._last_event_ts != ts:
            self._last_event_ts = ts
            self._last_second_events.clear()
            self._last_second_event_hashes = None
        self._last_second_events.append(event)

    @defer.inlineCallbacks
    def _update_last_event_ts(self) -> InlineCallbacksType[None]:
        if self._last_event_ts != self._last_event_ts_saved:
            self._last_event_ts_saved = self._last_event_ts
            yield self.master.db.state.setState(self._oid, "last_event_ts", self._last_event_ts)

    @defer.inlineCallbacks
    def _line_received_stream(self, line: bytes) -> InlineCallbacksType[None]:
        assert self._poll_handler is not None
        assert self._stream_activity_watchdog is not None

        self._stream_activity_watchdog.notify()
        try:
            event = json.loads(bytes2unicode(line))
        except ValueError:
            log.msg(f"{self.name}: bad json line: {line!r}")
            return

        if not is_event_valid(event):
            if self.debug:
                log.msg(f"no type in event {line!r}")
            return

        if not self._is_synchronized:
            self._queued_stream_events.append((extract_gerrit_event_time(event), event))
            if self._poll_connector is not None:
                self._poll_handler.force()
            return

        if self._stream_messages_timeout:
            self._stream_activity_watchdog.start()
        self._stream_messages_timeout = False
        self._poll_handler.stop()

        self._record_last_second_event(event, extract_gerrit_event_time(event))
        yield self._update_last_event_ts()
        yield self.eventReceived(event)

    def _filter_out_already_received_events(
        self, events: list[tuple[int, dict]]
    ) -> list[tuple[int, dict]]:
        if self._last_event_ts is None:
            return events

        filtered_events = []
        for ts, event in events:
            if ts < self._last_event_ts:
                continue
            if ts == self._last_event_ts:
                if self._last_second_event_hashes is not None:
                    if build_gerrit_event_hash(event) in self._last_second_event_hashes:
                        continue

                if event in self._last_second_events:
                    continue
            filtered_events.append((ts, event))
        return filtered_events

    def _debug_log_polled_event(self, event: dict) -> None:
        line = json.dumps(event, sort_keys=True)
        log.msg(f"{self.name} accepted polled event: {line}")

    @defer.inlineCallbacks
    def _lines_received_poll(self, lines: list[bytes]) -> InlineCallbacksType[None]:
        assert self._stream_connector is not None
        assert self._poll_handler is not None
        assert self._stream_activity_watchdog is not None

        if self._is_synchronized and not self._stream_messages_timeout:
            return

        # The code below parses all retrieved events, ignores already received ones, submits the
        # rest for processing and if it is detected that events from polling source are synchronized
        # with queued events from streaming source, switches to the streaming source.

        events = []
        for line in lines:
            try:
                event = json.loads(bytes2unicode(line))
            except ValueError:
                log.msg(f"{self.name}: bad json line: {line!r}")
                continue

            if not is_event_valid(event):
                if self.debug:
                    log.msg(f"no type in event {line!r}")
                continue
            events.append((extract_gerrit_event_time(event), event))

        events = sorted(self._filter_out_already_received_events(events), key=lambda e: e[0])

        if not events:
            self._poll_handler.schedule(self._http_poll_interval, invoke_again_if_running=True)
            return

        max_event_ts = events[-1][0]

        got_newer_messages_unhandled_before = True
        if self._last_event_ts is not None:
            got_newer_messages_unhandled_before = max_event_ts > self._last_event_ts

        needs_stream_restart = False
        if self._stream_messages_timeout and got_newer_messages_unhandled_before:
            # Stream connector has broken, because usually messages would come through the stream
            # connector first. Now new messages were received through HTTP API.
            #
            # Note: there must not be any asynchronous code between this check and the start of
            # the function to avoid race conditions.
            self._is_synchronized = False
            self._stream_messages_timeout = False
            needs_stream_restart = True

        if not self._queued_stream_events or max_event_ts <= self._queued_stream_events[0][0]:
            # The events from poll source has not caught up to stream events - process all events
            # and leave _is_synchronized as False.

            for ts, event in events:
                self._record_last_second_event(event, ts)
                if self.debug:
                    self._debug_log_polled_event(event)
                yield self.eventReceived(event)

            yield self._update_last_event_ts()
            self._poll_handler.schedule(self._http_poll_interval, invoke_again_if_running=True)
            if needs_stream_restart:
                self._deferwaiter.add(self._stream_connector.restart())
                self._stream_activity_watchdog.start()
            return

        first_queued_ts = self._queued_stream_events[0][0]

        # max_event_ts > first_queued_ts which means that:
        # - events list is guaranteed to have all events up to first_queued_ts (inclusive)
        # - starting with first_queued_ts (exclusive) the stream source has all events.
        for ts, event in events:
            if ts <= first_queued_ts:
                self._record_last_second_event(event, ts)
                if self.debug:
                    self._debug_log_polled_event(event)
                yield self.eventReceived(event)

        i = 0
        while i < len(self._queued_stream_events):
            ts, event = self._queued_stream_events[i]
            if ts == self._last_event_ts and event in self._last_second_events:
                i += 1
                continue
            self._record_last_second_event(event, ts)
            yield self.eventReceived(event)
            i += 1

        self._queued_stream_events.clear()
        if not needs_stream_restart:
            self._is_synchronized = True
        yield self._update_last_event_ts()
        if needs_stream_restart:
            self._deferwaiter.add(self._stream_connector.restart())
            self._stream_activity_watchdog.start()

    def describe(self) -> str:
        status = ""
        if not self._stream_connector or not self._stream_connector._process:
            status = "[NOT CONNECTED - check log]"
        return (
            "GerritChangeSource watching the remote "
            f"Gerrit repository {self.username}@{self.gerritserver} {status}"
        )


class GerritEventLogPoller(GerritChangeSourceBase):
    POLL_INTERVAL_SEC = 30
    FIRST_FETCH_LOOKBACK_DAYS = 30

    def checkConfig(  # type: ignore[override]
        self,
        baseURL: str,
        auth: Any,
        pollInterval: int = POLL_INTERVAL_SEC,
        pollAtLaunch: bool = True,
        firstFetchLookback: int = FIRST_FETCH_LOOKBACK_DAYS,
        **kwargs: Any,
    ) -> None:
        if self.name is None:
            self.name = f"GerritEventLogPoller:{baseURL}"
        super().checkConfig(**kwargs)

    @defer.inlineCallbacks
    def reconfigService(  # type: ignore[override]
        self,
        baseURL: str,
        auth: Any,
        pollInterval: int = POLL_INTERVAL_SEC,
        pollAtLaunch: bool = True,
        firstFetchLookback: int = FIRST_FETCH_LOOKBACK_DAYS,
        **kwargs: Any,
    ) -> InlineCallbacksType[None]:
        yield super().reconfigService(**kwargs)

        self._poll_interval = pollInterval
        self._poll_at_launch = pollAtLaunch

        self._oid = yield self.master.db.state.getObjectId(self.name, self.__class__.__name__)

        def get_last_event_ts() -> defer.Deferred:
            return self.master.db.state.getState(self._oid, 'last_event_ts', None)

        self._connector = GerritHttpEventLogPollerConnector(
            self.master.reactor,
            self,
            baseURL,
            auth,
            get_last_event_ts,
            first_fetch_lookback=firstFetchLookback,
            on_lines_received_cb=self._lines_received,
        )
        yield self._connector.setup()
        self._poller = util.poll.Poller(self._connector.do_poll, self, self.master.reactor)

    def getFiles(self, change: str, patchset: str) -> defer.Deferred:
        return self._connector.get_files(change, patchset)

    def force(self) -> None:
        self._poller()

    def activate(self) -> defer.Deferred[None]:
        self._poller.start(interval=self._poll_interval, now=self._poll_at_launch)
        return defer.succeed(None)

    def deactivate(self) -> defer.Deferred:
        return self._poller.stop()

    def describe(self) -> str:
        msg = "GerritEventLogPoller watching the remote Gerrit repository {}"
        return msg.format(self.name)

    @defer.inlineCallbacks
    def _lines_received(self, lines: list[bytes]) -> InlineCallbacksType[None]:
        last_event_ts = None
        for line in lines:
            try:
                event = json.loads(bytes2unicode(line))
            except ValueError:
                log.msg(f"{self.name}: bad json line: {line!r}")
                continue

            if not is_event_valid(event):
                if self.debug:
                    log.msg(f"no type in event {line!r}")
                continue

            yield super().eventReceived(event)

            this_last_event_ts = extract_gerrit_event_time(event)
            if last_event_ts is None:
                last_event_ts = this_last_event_ts
            else:
                last_event_ts = max(last_event_ts, this_last_event_ts)

        if last_event_ts is not None:
            yield self.master.db.state.setState(self._oid, "last_event_ts", last_event_ts)



#!/usr/bin/env python

import sys
import urllib

from future.utils import lrange

"""check_buildbot.py -H hostname -p httpport [options]

nagios check for buildbot.

requires that both metrics and web status enabled.

Both hostname and httpport must be set, or alternatively use url which
should be the full url to the metrics json resource"""
try:
    import simplejson as json
except ImportError:
    import json


OK, WARNING, CRITICAL, UNKNOWN = lrange(4)
STATUS_TEXT = ["OK", "Warning", "Critical", "Unknown"]
STATUS_CODES = dict(OK=OK, WARNING=WARNING, CRIT=CRITICAL)


def exit(level, msg):
    print(f"{STATUS_TEXT[level]}: {msg}")
    sys.exit(level)


def main():
    from optparse import OptionParser

    parser = OptionParser(__doc__)
    parser.set_defaults(hostname=None, httpport=None, url=None, verbosity=0)
    parser.add_option("-H", "--host", dest="hostname", help="Hostname")
    parser.add_option("-p", "--port", dest="httpport", type="int", help="WebStatus port")
    parser.add_option("-u", "--url", dest="url", help="Metrics url")
    parser.add_option(
        "-v", "--verbose", dest="verbosity", action="count", help="Increase verbosity"
    )
    options, args = parser.parse_args()

    if options.hostname and options.httpport:
        url = f"http://{options.hostname}:{options.httpport}/json/metrics"
    elif options.url:
        url = options.url
    else:
        exit(UNKNOWN, "You must specify both hostname and httpport, or just url")

    try:
        data = urllib.urlopen(url).read()
    except Exception:
        exit(CRITICAL, f"Error connecting to {url}")

    try:
        data = json.loads(data)
    except ValueError:
        exit(CRITICAL, f"Could not parse output of {url} as json")

    if not data:
        exit(WARNING, f"{url} returned null; are metrics disabled?")

    alarms = data['alarms']
    status = OK
    messages = []
    for alarm_name, alarm_state in alarms.items():
        if options.verbosity >= 2:
            messages.append(f"{alarm_name}: {alarm_state}")

        try:
            alarm_code = STATUS_CODES[alarm_state[0]]
        except (KeyError, IndexError):
            status = UNKNOWN
            messages.append(f"{alarm_name} has unknown alarm state {alarm_state}")
            continue

        status = max(status, alarm_code)
        if alarm_code > OK and options.verbosity < 2:
            messages.append(f"{alarm_name}: {alarm_state}")

    if not messages and status == OK:
        messages.append("no problems")
    exit(status, ";".join(messages))


if __name__ == '__main__':
    main()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.python import log

from buildbot.steps.source.git import Git


class GitLab(Git):
    """
    Source step that knows how to handle merge requests from
    the GitLab change source
    """

    def run_vc(self, branch, revision, patch):
        self.setup_repourl()
        # If this is a merge request:
        if self.build.hasProperty("target_branch"):
            target_repourl = self.build.getProperty("target_git_ssh_url", None)

            # repourl always includes ssh://
            if not target_repourl.startswith('ssh://'):
                target_repourl = 'ssh://' + target_repourl

            if self.repourl != target_repourl:
                log.msg(
                    "GitLab.run_vc: note: GitLab step for merge requests"
                    f" should probably have repourl='{target_repourl}' instead "
                    f"of '{self.repourl}'?"
                )
            # This step is (probably) configured to fetch the target
            # branch of a merge (because it is impractical for users to
            # configure one builder for each of the infinite number of
            # possible source branches for merge requests).
            # Point instead to the source being proposed for merge.
            branch = self.build.getProperty("source_branch", None)
            # FIXME: layering violation, should not be modifying self here?
            self.repourl = self.build.getProperty("source_git_ssh_url", None)
            # The revision is unlikely to exist in the repo already,
            # so tell Git to not check.
            revision = None

        return super().run_vc(branch, revision, patch)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


class Watchdog:
    def __init__(self, reactor, fn, timeout):
        self._reactor = reactor
        self._fn = fn
        self._timeout = timeout
        self._delayed_call = None

    def _timed_out(self):
        self._delayed_call = None
        self._fn()

    def start(self):
        if self._delayed_call is None:
            self._delayed_call = self._reactor.callLater(self._timeout, self._timed_out)

    def stop(self):
        if self._delayed_call is not None:
            self._delayed_call.cancel()

    def notify(self):
        if self._delayed_call is None:
            return

        self._delayed_call.reset(self._timeout)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.python import log

from buildbot import config
from buildbot.statistics.storage_backends.base import StatsStorageBase

try:
    from influxdb import InfluxDBClient
except ImportError:
    InfluxDBClient = None


class InfluxStorageService(StatsStorageBase):
    """
    Delegates data to InfluxDB
    """

    def __init__(self, url, port, user, password, db, captures, name="InfluxStorageService"):
        if not InfluxDBClient:
            config.error("Python client for InfluxDB not installed.")
            return
        self.url = url
        self.port = port
        self.user = user
        self.password = password
        self.db = db
        self.name = name

        self.captures = captures
        self.client = InfluxDBClient(self.url, self.port, self.user, self.password, self.db)
        self._inited = True

    def thd_postStatsValue(self, post_data, series_name, context=None):
        if not self._inited:
            log.err(f"Service {self.name} not initialized")
            return

        data = {'measurement': series_name, 'fields': post_data}

        log.msg("Sending data to InfluxDB")
        log.msg(f"post_data: {post_data!r}")
        if context:
            log.msg(f"context: {context!r}")
            data['tags'] = context

        self.client.write_points([data])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process import buildstep
from buildbot.process.results import SUCCESS
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import TestBuildStepMixin
from buildbot.test.util.warnings import assertProducesWarning
from buildbot.test.util.warnings import assertProducesWarnings
from buildbot.warnings import DeprecatedApiWarning


class TestStep(buildstep.ShellMixin, buildstep.BuildStep):
    def __init__(self, text):
        self.setupShellMixin({})
        super().__init__()
        self.text = text

    @defer.inlineCallbacks
    def run(self):
        for file in self.build.allFiles():
            cmd = yield self.makeRemoteShellCommand(command=["echo", "build_file", file])
            yield self.runCommand(cmd)
        version = self.build.getWorkerCommandVersion("shell", None)
        if version != "99.99":
            cmd = yield self.makeRemoteShellCommand(command=["echo", "version", version])
            yield self.runCommand(cmd)
        cmd = yield self.makeRemoteShellCommand(command=["echo", "done", self.text])
        yield self.runCommand(cmd)
        return SUCCESS


class TestTestBuildStepMixin(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    @defer.inlineCallbacks
    def test_setup_build(self):
        self.setup_build(
            worker_version={"*": "2.9"}, worker_env={"key": "value"}, build_files=["build.txt"]
        )
        self.setup_step(TestStep("step1"))
        self.expect_commands(
            ExpectShell(
                workdir="wkdir", command=["echo", "build_file", "build.txt"], env={"key": "value"}
            ).exit(0),
            ExpectShell(
                workdir="wkdir", command=["echo", "version", "2.9"], env={"key": "value"}
            ).exit(0),
            ExpectShell(
                workdir="wkdir", command=["echo", "done", "step1"], env={"key": "value"}
            ).exit(0),
        )
        self.expect_outcome(SUCCESS)
        yield self.run_step()

    @defer.inlineCallbacks
    def test_old_setup_step_args(self):
        with assertProducesWarnings(
            DeprecatedApiWarning,
            num_warnings=3,
            message_pattern=".*has been deprecated, use setup_build\\(\\) to pass this information",
        ):
            self.setup_step(
                TestStep("step1"),
                worker_version={"*": "2.9"},
                worker_env={"key": "value"},
                build_files=["build.txt"],
            )
        self.expect_commands(
            ExpectShell(
                workdir="wkdir", command=["echo", "build_file", "build.txt"], env={"key": "value"}
            ).exit(0),
            ExpectShell(
                workdir="wkdir", command=["echo", "version", "2.9"], env={"key": "value"}
            ).exit(0),
            ExpectShell(
                workdir="wkdir", command=["echo", "done", "step1"], env={"key": "value"}
            ).exit(0),
        )
        self.expect_outcome(SUCCESS)

        yield self.run_step()

    def test_get_nth_step(self):
        self.setup_step(TestStep("step1"))
        self.assertTrue(isinstance(self.get_nth_step(0), TestStep))

        with assertProducesWarning(DeprecatedApiWarning, "step attribute has been deprecated"):
            self.assertTrue(isinstance(self.step, TestStep))

    @defer.inlineCallbacks
    def test_multiple_steps(self):
        self.setup_step(TestStep("step1"))
        self.setup_step(TestStep("step2"))
        self.expect_commands(
            ExpectShell(workdir="wkdir", command=["echo", "done", "step1"]).stdout("out1").exit(0),
            ExpectShell(workdir="wkdir", command=["echo", "done", "step2"]).stdout("out2").exit(0),
        )
        self.expect_log_file("stdio", "out1\n", step_index=0)
        self.expect_log_file("stdio", "out2\n", step_index=1)
        self.expect_outcome(SUCCESS)
        self.expect_outcome(SUCCESS)
        yield self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from buildbot.data import base
from buildbot.data import types
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    from buildbot.db.codebase_branches import CodebaseBranchModel


def branch_db_to_data(model: CodebaseBranchModel) -> dict[str, Any]:
    return {
        "branchid": model.id,
        "codebaseid": model.codebaseid,
        "name": model.name,
        "commitid": model.commitid,
        "last_timestamp": model.last_timestamp,
    }


branches_field_map = {
    'branchid': 'codebase_branches.id',
    'codebaseid': 'codebase_branches.codebaseid',
    'name': 'codebase_branches.name',
    'commitid': 'codebase_branches.commitid',
    'last_timestamp': 'codebase_branches.last_timestamp',
}


class CodebaseBranchEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/branches/n:branchid",
    ]

    @async_to_deferred
    async def get(self, result_spec: base.ResultSpec, kwargs: Any) -> dict[str, Any] | None:
        branch = await self.master.db.codebase_branches.get_branch(kwargs['branchid'])
        if not branch:
            return None
        return branch_db_to_data(branch)


class CodebaseBranchesEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/codebases/n:codebaseid/branches",
    ]

    @async_to_deferred
    async def get(self, result_spec: base.ResultSpec, kwargs: Any) -> list[dict[str, Any]]:
        result_spec.fieldMapping = branches_field_map
        branches = await self.master.db.codebase_branches.get_branches(
            kwargs['codebaseid'], result_spec=result_spec
        )
        return [branch_db_to_data(c) for c in branches]


class CodebaseBranch(base.ResourceType):
    name = "branch"
    plural = "branches"
    endpoints = [CodebaseBranchEndpoint, CodebaseBranchesEndpoint]
    eventPathPatterns = [
        "/branches/:branchid",
    ]

    class EntityType(types.Entity):
        branchid = types.Integer()
        codebaseid = types.Integer()
        name = types.String()
        commitid = types.Integer()
        last_timestamp = types.Integer()

    entityType = EntityType(name)

    @async_to_deferred
    async def generate_event(self, _id: int, event: str) -> None:
        branch = await self.master.data.get(('branches', _id))
        self.produceEvent(branch, event)

    @base.updateMethod
    @async_to_deferred
    async def update_branch(
        self,
        *,
        codebaseid: int,
        name: str,
        commitid: int | None = None,
        last_timestamp: int,
    ) -> None:
        await self.master.db.codebase_branches.update_branch(
            codebaseid=codebaseid,
            name=name,
            commitid=commitid,
            last_timestamp=last_timestamp,
        )
        branch = await self.master.db.codebase_branches.get_branch_by_name(
            codebaseid=codebaseid, name=name
        )
        assert branch is not None
        await self.generate_event(branch.id, 'update')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from typing import ClassVar
from typing import Sequence

from twisted.internet import defer
from zope.interface import implementer

from buildbot import config
from buildbot import interfaces
from buildbot import util
from buildbot.reporters.message import MessageFormatterMissingWorker

ENCODING = 'utf-8'


@implementer(interfaces.IReportGenerator)
class WorkerMissingGenerator(util.ComparableMixin):
    compare_attrs: ClassVar[Sequence[str]] = ['workers', 'formatter']

    wanted_event_keys = [
        ('workers', None, 'missing'),
    ]

    def __init__(self, workers='all', message_formatter=None):
        self.workers = workers
        self.formatter = message_formatter
        if self.formatter is None:
            self.formatter = MessageFormatterMissingWorker()

    def check(self):
        if not (self.workers == 'all' or isinstance(self.workers, (list, tuple, set))):
            config.error("workers must be 'all', or list of worker names")

    @defer.inlineCallbacks
    def generate(self, master, reporter, key, worker):
        if not self._is_message_needed(worker):
            return None

        msg = yield self.formatter.formatMessageForMissingWorker(master, worker)
        body = msg['body'].encode(ENCODING)
        subject = msg['subject']
        if subject is None:
            subject = f"Buildbot worker {worker['name']} missing"
        assert msg['type'] in (
            'plain',
            'html',
        ), f"'{msg['type']}' message type must be 'plain' or 'html'."

        return {
            'body': body,
            'subject': subject,
            'type': msg['type'],
            'results': None,
            'builds': None,
            "buildset": None,
            'users': worker['notify'],
            'patches': None,
            'logs': None,
            'worker': worker['name'],
        }

    def generate_name(self):
        name = self.__class__.__name__
        if self.workers is not None:
            name += "_workers_" + "+".join(self.workers)
        return name

    def _is_message_needed(self, worker):
        return (self.workers == 'all' or worker['name'] in self.workers) and worker['notify']



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class Builder(Row):
    table = "builders"

    id_column = 'id'
    hashedColumns = [('name_hash', ('name',))]

    def __init__(
        self,
        id=None,
        name=None,
        name_hash=None,
        projectid=None,
        description=None,
        description_format=None,
        description_html=None,
    ):
        if name is None:
            name = f'builder-{id}'
        super().__init__(
            id=id,
            name=name,
            name_hash=name_hash,
            projectid=projectid,
            description=description,
            description_format=description_format,
            description_html=description_html,
        )


class BuilderMaster(Row):
    table = "builder_masters"
    id_column = 'id'

    def __init__(self, id=None, builderid=None, masterid=None):
        super().__init__(id=id, builderid=builderid, masterid=masterid)


class BuildersTags(Row):
    table = "builders_tags"
    id_column = 'id'

    def __init__(self, id=None, builderid=None, tagid=None):
        super().__init__(id=id, builderid=builderid, tagid=tagid)
